"""BLACS distributed matrix object."""
from __future__ import annotations

from types import ModuleType
from typing import Dict, Tuple
import gpaw.cgpaw as cgpaw
import numpy as np
import scipy.linalg as sla

import gpaw.utilities.blas as blas
from gpaw import debug, get_scipy_version
from gpaw.gpu import cupy as cp, cupy_eigh, XP
from gpaw.mpi import MPIComm, _Communicator, serial_comm
from gpaw.typing import Array1D, ArrayLike1D, ArrayLike2D, Array2D

_global_blacs_context_store: Dict[Tuple[_Communicator, int, int], int] = {}


def suggest_blocking(N: int, ncpus: int) -> tuple[int, int, int]:
    """Suggest blocking of ``NxN`` matrix.

    Returns rows, columns, blocksize tuple.

    >>> suggest_blocking(10, 6)
    (3, 2, 2)
    """

    nprow = ncpus
    npcol = 1

    # Make npcol and nprow as close to each other as possible
    npcol_try = npcol
    while npcol_try < nprow:
        if ncpus % npcol_try == 0:
            npcol = npcol_try
            nprow = ncpus // npcol
        npcol_try += 1

    assert npcol * nprow == ncpus

    # ScaLAPACK creates trouble if there aren't at least a few whole blocks.
    # Choose block size so that there will always be at least one whole block
    # and at least two blocks in total.
    blocksize = max((N - 2) // max(nprow, npcol), 1)
    # The next commented line would give more whole blocks.
    # blocksize = max(N // max(nprow, npcol) - 2, 1)

    # Use block size that is a power of 2 and at most 64
    blocksize = 2**int(np.log2(blocksize))
    blocksize = max(min(blocksize, 64), 1)

    return nprow, npcol, blocksize


class Matrix(XP):
    def __init__(self,
                 M: int,
                 N: int,
                 dtype=None,
                 data: ArrayLike2D | None = None,
                 dist: MatrixDistribution | tuple | None = None,
                 xp=None):
        """Matrix object.

        Parameters
        ----------
        M:
            Rows.
        N:
            Columns.
        dtype:
            Data type (float or complex).
        dist:
            BLACS distribution given as
            (communicator, rows, columns, blocksize)
            tuple.  Default is None meaning no distribution.
        data:
            Numpy ndarray to use for storage.  By default, a new ndarray
            will be allocated.
            """
        self.shape = (M, N)

        if data is None or isinstance(data, (np.ndarray, cp.ndarray)):
            pass
        else:
            data = np.asarray(data)

        if dtype is None:
            if data is None:
                dtype = float
            else:
                dtype = data.dtype
        self.dtype = np.dtype(dtype)
        assert dtype == float or dtype == complex, dtype

        self.xp: ModuleType
        if xp is None:
            if isinstance(dist, CuPyDistribution):
                xp = cp
            elif data is not None and not isinstance(data, np.ndarray):
                xp = cp
            else:
                xp = np
        XP.__init__(self, xp)

        dist = dist or ()
        if isinstance(dist, tuple):
            kwargs = {key: val for key, val in zip(['comm', 'r', 'c', 'b'],
                                                   dist)}
            dist = create_distribution(M, N, xp=self.xp, **kwargs)
        else:
            assert self.shape == dist.full_shape
        self.dist = dist

        self.data: Array2D
        if data is None:
            self.data = self.xp.empty(dist.shape, self.dtype)
        else:
            assert data.shape == dist.shape, (data.shape, dist.shape, dist)
            self.data = data

    def __repr__(self):
        dist = str(self.dist).split('(')[1]
        if self.xp is cp:
            dist = 'xp=cp, ' + dist
        return f'Matrix({self.dtype.name}: {dist}'

    def new(self, dist='inherit', data=None) -> Matrix:
        """Create new matrix of same shape and dtype.

        Default is to use same BLACS distribution.  Use dist to use another
        distribution.
        """
        return Matrix(*self.shape,
                      dtype=self.dtype,
                      dist=self.dist if dist == 'inherit' else dist,
                      data=data,
                      xp=self.xp)

    def copy(self) -> Matrix:
        """Create a copy."""
        M = self.new()
        M.data[:] = self.data
        return M

    def __setitem__(self, item, value):
        assert item == slice(None)
        assert isinstance(value, Matrix)
        self.data[:] = value.data

    def __iadd__(self, other):
        if isinstance(other, Matrix):
            other = other.data
        self.data += other
        return self

    def multiply(self,
                 other,
                 alpha=1.0,
                 opa='N',
                 opb='N',
                 out=None,
                 beta=0.0,
                 symmetric=False) -> Matrix:
        """BLAS matrix-multiplication with other matrix."""
        if not isinstance(other, Matrix):
            other = other.matrix
        A = self
        B = other
        dist = self.dist
        if out is None:
            assert beta == 0.0
            M = A.shape[0] if opa == 'N' else A.shape[1]
            N = B.shape[1] if opb == 'N' else B.shape[0]
            out = Matrix(M, N, A.dtype, dist=dist.new(M, N))
        elif not isinstance(out, Matrix):
            out = out.matrix

        dist.multiply(alpha, A, opa, B, opb, beta, out, symmetric=symmetric)
        return out

    def redist(self, other: Matrix) -> None:
        """Redistribute to other BLACS layout."""
        if self is other:
            return
        d1 = self.dist
        d2 = other.dist
        n1 = d1.rows * d1.columns
        n2 = d2.rows * d2.columns
        if n1 == n2 == 1:
            other.data[:] = self.data
            return

        if n2 == 1 and d1.blocksize is None:
            assert d2.blocksize is None
            assert d1.columns == 1
            comm = d1.comm
            if comm.rank == 0:
                M = self.shape[0]
                m = (M + comm.size - 1) // comm.size
                other.data[:m] = self.data
                for r in range(1, comm.size):
                    m1 = min(r * m, M)
                    m2 = min(m1 + m, M)
                    comm.receive(other.data[m1:m2], r)
            else:
                comm.send(self.data, 0)
            return

        if n1 == 1 and d2.blocksize is None:
            assert d1.blocksize is None
            assert d1.columns == 1
            comm = d1.comm
            if comm.rank == 0:
                M = self.shape[0]
                m = (M + comm.size - 1) // comm.size
                other.data[:] = self.data[:m]
                for r in range(1, comm.size):
                    m1 = min(r * m, M)
                    m2 = min(m1 + m, M)
                    comm.send(self.data[m1:m2], r)
            else:
                comm.receive(other.data, 0)
            return

        c = d1.comm if d1.comm.size > d2.comm.size else d2.comm
        n = max(n1, n2)
        if n < c.size:
            c = c.new_communicator(np.arange(n))
        if c is not None:
            M, N = self.shape
            d1 = create_distribution(M, N, c,
                                     d1.rows, d1.columns, d1.blocksize)
            d2 = create_distribution(M, N, c,
                                     d2.rows, d2.columns, d2.blocksize)
            if n1 == n:
                ctx = d1.desc[1]
            else:
                ctx = d2.desc[1]
            redist(d1, self.data, d2, other.data, ctx)

    def gather(self, root: int = 0, broadcast=False) -> Matrix:
        """Gather the Matrix on the root rank.

        Returns a new Matrix distributed so that all data is on the root rank
        """
        assert root == 0
        if self.dist.comm.size > 1:
            S = self.new(dist=(self.dist.comm, 1, 1))
            self.redist(S)
            if broadcast:
                if self.dist.comm.rank > 0:
                    S = self.new(dist=None)
                self.dist.comm.broadcast(S.data, 0)
        else:
            S = self

        return S

    def inv(self, uplo='L'):
        """Inplace inversion."""
        assert uplo == 'L'
        M, N = self.shape
        assert M == N
        dist = self.dist
        if dist.comm.size == 1:
            self.tril2full()
            self.data[:] = sla.inv(self.data,
                                   overwrite_a=True,
                                   check_finite=debug)
            return
        bc, br = dist.desc[4:6]
        assert bc == br
        info = cgpaw.scalapack_inverse(self.data, dist.desc, 'U')
        if info != 0:
            raise ValueError(f'scalapack_inverse error: {info}')

    def invcholesky(self) -> None:
        """In-place inverse of Cholesky decomposition.

        Calculate a lower triangle matrix `L` where:::

             †
          LSL = 1,

        and `S` is self.  Only the lower part of `S` is used.

        >>> S = Matrix(2, 2, data=[[1.0, np.nan],
        ...                        [0.1, 1.0]])
        >>> S.invcholesky()
        >>> S.data
        array([[ 1.        , -0.        ],
               [-0.10050378,  1.00503782]])
        """
        S = self.gather()
        if self.dist.comm.rank == 0:
            if isinstance(S.data, np.ndarray):
                if debug:
                    S.data[np.triu_indices(S.shape[0], 1)] = 42.0
                L_nn = sla.cholesky(S.data,
                                    lower=True,
                                    overwrite_a=True,
                                    check_finite=debug)
                S.data[:] = sla.inv(L_nn,
                                    overwrite_a=True,
                                    check_finite=debug)
            else:
                S.tril2full()
                L_nn = cp.linalg.cholesky(S.data)
                S.data[:] = cp.linalg.inv(L_nn)

        if S is not self:
            S.redist(self)

    def eigh(self,
             S=None,
             *,
             cc=False,
             scalapack=(None, 1, 1, None),
             limit: int | None = None) -> Array1D:
        """Calculate eigenvectors and eigenvalues.

        Matrix must be symmetric/hermitian and stored in lower half.
        If ``S`` is given, solve a generalized eigenvalue problem.

        Parameters
        ----------
        cc: bool
            Complex conjugate matrix before finding eigenvalues.
        scalapack: tuple
            BLACS distribution for ScaLapack to use.  Default is to do serial
            diagonalization.
        limit:
            Number of eigenvector and values to find.  Defaults to all.
        """
        slcomm, rows, columns, blocksize = scalapack
        slcomm = slcomm or self.dist.comm
        dist = (slcomm, rows, columns, blocksize)

        redist = (rows != self.dist.rows or
                  columns != self.dist.columns or
                  blocksize != self.dist.blocksize)

        if redist:
            H = self.new(dist=dist)
            self.redist(H)
            if S is not None:
                S0 = S
                S = S0.new(dist=dist)
                S0.redist(S)
        else:
            assert self.dist.comm.size == slcomm.size
            H = self

        if limit:
            eps = self.xp.empty(limit)
        else:
            eps = self.xp.empty(H.shape[0])

        if rows * columns == 1:
            if self.dist.comm.rank == 0:
                if cc and H.dtype == complex:
                    np.negative(H.data.imag, H.data.imag)
                if debug:
                    H.data[np.triu_indices(H.shape[0], 1)] = 42.0
                if S is None:
                    if self.xp is not np:
                        assert isinstance(H.data, cp.ndarray)
                        eps[:], H.data.T[:] = cupy_eigh(H.data, UPLO='L')
                    else:
                        eps[:], H.data.T[:] = sla.eigh(
                            H.data,
                            lower=True,
                            overwrite_a=True,
                            check_finite=debug,
                            driver='evx' if H.data.size == 1 else 'evd')
                else:
                    if self.xp is cp:
                        assert self.dist.comm.size == 1
                        S.invcholesky()
                        self.tril2full()
                        eigs = self.eighg(S)
                        self.data[:] = self.data.T.copy()
                        return eigs
                    if debug:
                        S.data[self.xp.triu_indices(H.shape[0], 1)] = 42.0
                    eps, evecs = sla.eigh(
                        H.data,
                        S.data,
                        lower=True,
                        overwrite_a=True,
                        overwrite_b=True,
                        check_finite=debug,
                        subset_by_index=(0, limit - 1) if limit else None)
                    limit = limit or len(eps)
                    H.data.T[:, :limit] = evecs
            self.dist.comm.broadcast(eps, 0)
        else:
            if slcomm.rank < rows * columns:
                assert cc
                assert S is None
                array = H.data.copy()
                info = cgpaw.scalapack_diagonalize_dc(array, H.dist.desc, 'U',
                                                      H.data, eps)
                assert info == 0, info

            # necessary to broadcast eps when some ranks are not used
            # in current scalapack parameter set
            # eg. (2, 1, 2) with 4 processes
            if rows * columns < slcomm.size:
                H.dist.comm.broadcast(eps, 0)

        if redist:
            H.redist(self)

        return eps

    def eighg(self, L: Matrix, comm2: MPIComm = serial_comm) -> Array1D:
        """Solve generalized eigenvalue problem.

        With `H` being self, we solve for the eigenvectors `C` and the
        eigenvalues `Λ` (a diagonal matrix):::

          HC = SCΛ,

        where `L` is a lower triangle matrix such that:::

             †
          LSL = 1.

        The solution has these three steps:::

           ~      †   ~~   ~         †~
           H = LHL ,  HC = CΛ,  C = L C.

        Note that `H` must be the full matrix not just half of it!

        """
        M, N = self.shape
        assert M == N
        comm = self.dist.comm

        if comm2.rank == 0:
            if comm.size == 1:
                H = self
                L0 = L
            else:
                # TODO: Use scalapack
                H = self.new(dist=(comm,))
                self.redist(H)
                L0 = self.new(dist=(comm,))
                L.redist(L0)
            if comm.rank == 0:
                if self.xp is not np:
                    return self.dist.eighg(self, L0)
                tmp_MM = np.empty_like(H.data)
                L_MM = L0.data
                blas.mmm(1.0, L_MM, 'N', H.data, 'N', 0.0, tmp_MM)
                blas.r2k(0.5, tmp_MM, L_MM, 0.0, H.data)
                # Ht_MM = L_MM @ H.data @ L_MM.conj().T
                if get_scipy_version() >= [1, 9]:
                    driver = 'evx' if M == 1 else 'evd'
                else:
                    driver = None
                eig_n, Ct_Mn = sla.eigh(
                    H.data,
                    overwrite_a=True,
                    check_finite=debug,
                    driver=driver)
                assert Ct_Mn.flags.f_contiguous
                blas.mmm(1.0, L_MM, 'C', Ct_Mn.T, 'T', 0.0, H.data)
                # H.data[:] = L_MM.T.conj() @ Ct_Mn
            else:
                eig_n = np.empty(M)

            if comm.size > 1:
                H.redist(self)
                comm.broadcast(eig_n, 0)

        if comm2.rank > 0:
            eig_n = np.empty(M)
        comm2.broadcast(eig_n, 0)
        comm2.broadcast(self.data, 0)

        return eig_n

    def complex_conjugate(self) -> None:
        """Inplace complex conjugation."""
        if self.dtype == complex:
            self.xp.negative(self.data.imag, self.data.imag)

    def add_hermitian_conjugate(self, scale: float = 1.0) -> None:
        """Add hermitian conjugate to myself."""
        if self.dist.comm.size == 1:
            if scale != 1.0:
                self.data *= scale
            self.data += self.data.conj().T
            return
        tmp = self.copy()
        cgpaw.pblas_tran(*self.shape, scale, tmp.data, scale, self.data,
                         self.dist.desc, self.dist.desc, True)

    def tril2full(self) -> None:
        """Fill in upper triangle from lower triangle.

        For a real matrix::

          a ? ?    a b d
          b c ? -> b c e
          d e f    d e f

        For a complex matrix, the complex conjugate of the lower part will
        be inserted into the upper part.
        """
        M, N = self.shape
        assert M == N

        dist = self.dist

        if dist.comm.size == 1 or dist.rows == 1 and dist.columns == 1:
            if dist.comm.rank == 0:
                lower = self.xp.tri(M, k=-1, dtype=bool)
                self.data.T[lower] = self.data[lower].conj()
            return

        desc = dist.desc
        cgpaw.scalapack_set(self.data, desc, 0.0, 0.0, 'L', M - 1, M - 1, 2, 1)
        buf = self.data.copy()
        # Set diagonal to zero in the copy:
        cgpaw.scalapack_set(buf, desc, 0.0, 0.0, 'L', M, M, 1, 1)
        # Now transpose tmp_mm adding the result to the original matrix:
        cgpaw.pblas_tran(M, M, 1.0, buf, 1.0, self.data, desc, desc, True)

    def add_to_diagonal(self, d: ArrayLike1D | float) -> None:
        """Add list of numbers or single number to diagonal of matrix."""
        n1, n2 = self.dist.my_row_range()
        M, N = self.shape
        assert M == N
        self.data.ravel()[n1::N + 1] += d

    def to_cpu(self):
        if isinstance(self.data, np.ndarray):
            return self
        return Matrix(*self.shape, data=cp.asnumpy(self.data))


def _matrix(M):
    """Dig out Matrix object from wrapper(s)."""
    if isinstance(M, Matrix):
        return M
    return _matrix(M.matrix)


def redist(dist1, M1, dist2, M2, context):
    cgpaw.scalapack_redist(dist1.desc, dist2.desc,
                           M1, M2,
                           dist1.desc[2], dist1.desc[3],
                           1, 1, 1, 1,  # 1-indexing
                           context, 'G')


def create_distribution(M: int,
                        N: int,
                        comm: MPIComm | None = None,
                        r: int = 1,
                        c: int = 1,
                        b: int | None = None,
                        xp=None) -> MatrixDistribution:
    if xp is cp:
        assert b is None
        if r == 1 and c == 1:
            pass  # comm = None
        comm = comm or serial_comm
        return CuPyDistribution(M, N, comm,
                                r if r != -1 else comm.size,
                                c if c != -1 else comm.size,
                                b)

    if comm is None or comm.size == 1:
        assert r == 1 and abs(c) == 1 or c == 1 and abs(r) == 1
        return NoDistribution(M, N)

    return BLACSDistribution(M, N, comm,
                             r if r != -1 else comm.size,
                             c if c != -1 else comm.size,
                             b)


class MatrixDistribution:
    comm: MPIComm
    rows: int
    columns: int
    blocksize: int | None  # None means everything on rank=0
    shape: tuple[int, int]
    full_shape: tuple[int, int]
    desc: Array1D

    def matrix(self, dtype=None, data=None):
        return Matrix(*self.full_shape, dtype=dtype, data=data, dist=self)

    def multiply(self, alpha, a, opa, b, opb, beta, c, symmetric):
        raise NotImplementedError

    def eighg(self, H, L):
        raise NotImplementedError

    def new(self, M, N):
        raise NotImplementedError

    def my_row_range(self) -> tuple[int, int]:
        """Return indices for range of my rows.

        >>> Matrix(2, 2).dist.my_row_range()
        (0, 2)
        """
        ok = (self.rows == self.comm.size and
              self.columns == 1 and
              self.blocksize is None)
        if not ok:
            raise ValueError(f'Can not create slice of distribution: {self}')
        M = self.full_shape[0]
        b = (M + self.rows - 1) // self.rows
        n1 = self.comm.rank * b
        n2 = min(n1 + b, M)
        return n1, n2


class NoDistribution(MatrixDistribution):
    comm = serial_comm
    rows = 1
    columns = 1
    blocksize = None

    def __init__(self, M, N):
        self.shape = (M, N)
        self.full_shape = (M, N)

    def __str__(self):
        return 'NoDistribution({}x{})'.format(*self.shape)

    def global_index(self, n):
        return n

    def new(self, M, N):
        return NoDistribution(M, N)

    def multiply(self, alpha, a, opa, b, opb, beta, c, symmetric):
        if symmetric:
            if opa == 'N':
                assert opb == 'C' or opb == 'T' and a.dtype == float
                if a is b:
                    blas.rk(alpha, a.data, beta, c.data)
                else:
                    if beta == 1.0 and a.shape[1] == 0:
                        return
                    blas.r2k(0.5 * alpha, a.data, b.data, beta, c.data)
            else:
                1 / 0
                assert opa == 'C' and opb == 'N'
                assert a is not b
                blas.r2k(0.5 * alpha, a.data, b.data, beta, c.data, 'n')

        else:
            blas.mmm(alpha, a.data, opa, b.data, opb, beta, c.data)


class BLACSDistribution(MatrixDistribution):
    serial = False

    def __init__(self, M, N, comm, r, c, b):
        self.comm = comm
        self.rows = r
        self.columns = c
        self.blocksize = b
        self.full_shape = (M, N)
        self.simple = False

        key = (comm, r, c)
        context = _global_blacs_context_store.get(key)
        if context is None:
            try:
                context = cgpaw.new_blacs_context(comm.get_c_object(),
                                                  c, r, 'R')
            except AttributeError:
                pass
            else:
                _global_blacs_context_store[key] = context

        if b is None:
            if c == 1:
                br = (M + r - 1) // r
                bc = max(1, N)
                self.simple = True
            elif r == 1:
                br = M
                bc = (N + c - 1) // c
            else:
                raise ValueError('Please specify block size!')
        else:
            br = bc = b

        if context is None:
            assert b is None
            assert c == 1
            n = N
            m = min((comm.rank + 1) * br, M) - min(comm.rank * br, M)
        else:
            n, m = cgpaw.get_blacs_local_shape(context, N, M, bc, br, 0, 0)
        if n < 0 or m < 0:
            n = m = 0
        self.shape = (m, n)
        lld = max(1, n)
        if context is not None:
            self.desc = np.array([1, context, N, M, bc, br, 0, 0, lld],
                                 np.intc)

    def __str__(self):
        return ('BLACSDistribution(global={}, local={}, blocksize={})'
                .format(*('{}x{}'.format(*shape)
                          for shape in [self.desc[3:1:-1],
                                        self.shape,
                                        self.desc[5:3:-1]])))

    def global_index(self, myi):
        return self.comm.rank * int(self.desc[5]) + myi

    def new(self, M, N):
        return BLACSDistribution(M, N,
                                 self.comm,
                                 self.rows, self.columns,
                                 self.blocksize)

    def multiply(self, alpha, a, opa, b, opb, beta, c, symmetric):
        if self.comm.size > 1:
            ok = a.dist.simple and b.dist.simple and c.dist.simple
            if ok:
                # Special cases that don't need scalapack - most likely also
                # faster:
                if opa == 'N' and opb == 'N':
                    return mmm_nn(a, b, c, alpha, beta, blas.mmm)
                if opa == 'N' and opb == 'C':
                    if symmetric:
                        if beta == 1.0:
                            return mmm_nc_sym(a, b, c, alpha, blas.mmm)
                    else:
                        return mmm_nc(a, b, c, alpha, beta, blas.mmm)

        if symmetric:
            assert opa == 'N'
            assert opb == 'C' or opb == 'T' and a.dtype == float
            N, K = a.shape
            if a is b:
                cgpaw.pblas_rk(N, K, alpha, a.data,
                               beta, c.data,
                               a.dist.desc, c.dist.desc,
                               'U')
            else:
                cgpaw.pblas_r2k(N, K, 0.5 * alpha, b.data, a.data,
                                beta, c.data,
                                b.dist.desc, a.dist.desc, c.dist.desc,
                                'U')
        else:
            Ka, M = a.shape
            N, Kb = b.shape
            if opa == 'N':
                Ka, M = M, Ka
            if opb == 'N':
                N, Kb = Kb, N
            cgpaw.pblas_gemm(N, M, Ka, alpha, b.data, a.data,
                             beta, c.data,
                             b.dist.desc, a.dist.desc, c.dist.desc,
                             opb, opa)


def cublas_mmm(alpha, a, opa, b, opb, beta, c):
    if c.size == 0:
        return
    if a.size == 0 and beta == 1.0:
        return
    cp.cublas.gemm(opa.replace('C', 'H'), opb.replace('C', 'H'),
                   a, b, c, alpha, beta)


class CuPyDistribution(MatrixDistribution):
    def __init__(self, M, N, comm, r, c, b):
        self.comm = comm
        self.rows = r
        self.columns = c
        self.blocksize = b
        self.full_shape = (M, N)
        # assert r == comm.size, (M, N, comm, r, c, b)
        assert c == 1
        br = (M + r - 1) // r
        m = min((comm.rank + 1) * br, M) - min(comm.rank * br, M)
        self.shape = (m, N)

    def __str__(self):
        M, N = self.full_shape
        m, N = self.shape
        return f'CuPyDistribution(global={M}x{N}, local={m}x{N})'

    def global_index(self, n):
        1 / 0
        return n

    def new(self, M, N):
        return CuPyDistribution(M, N,
                                self.comm,
                                self.rows, self.columns,
                                self.blocksize)

    def multiply(self, alpha, a, opa, b, opb, beta, c, *, symmetric=False):
        if self.comm.size > 1:
            if opa == 'N' and opb == 'N':
                return mmm_nn(a, b, c, alpha, beta, cublas_mmm)
            if opa == 'N' and opb == 'C':
                if symmetric:
                    if beta == 1.0:
                        return mmm_nc_sym(a, b, c, alpha, cublas_mmm)
                else:
                    return mmm_nc(a, b, c, alpha, beta, cublas_mmm)
            1 / 0

        if symmetric:
            if opa == 'N':
                assert opb == 'C' or opb == 'T' and a.dtype == float
                if a is b:
                    cp.cublas.gemm('N', 'H',
                                   a.data, a.data, c.data,
                                   alpha, beta)
                    # cp.cublas.syrk('N', a.data, c.data, alpha, beta, True)
                else:
                    if beta == 1.0 and a.shape[1] == 0:
                        return
                    if c.data.size > 0:
                        assert beta in [0.0, 1.0]
                        cp.cublas.gemm('N', 'H',
                                       a.data, b.data, c.data,
                                       alpha, beta)
            else:
                1 / 0
                assert opa == 'C' and opb == 'N'
                assert a is not b
                raise NotImplementedError
                blas.r2k(0.5 * alpha, a.data, b.data, beta, c.data, 'n')

        else:
            cublas_mmm(alpha, a.data, opa, b.data, opb, beta, c.data)

    def eighg(self, H, L):
        """
        :::

           ~      †   ~~   ~         †~
           H = LHL ,  HC = CΛ,  C = L C.
        """
        assert self.comm.size == 1
        tmp = H.new()
        self.multiply(1.0, L, 'N', H, 'N', 0.0, tmp)
        self.multiply(1.0, tmp, 'N', L, 'C', 0.0, H, symmetric=True)
        eig_M, Ct_MM = cupy_eigh(H.data, UPLO='L')
        assert Ct_MM.flags.f_contiguous
        Ct = H.new(data=Ct_MM.T)
        self.multiply(1.0, L, 'C', Ct, 'T', 0.0, H)
        # H.complex_conjugate()
        return eig_M


def mmm_nn(m1, m2, m3, alpha, beta, mmm):
    """Parallel matrix-matrix multiplication.

    :::

        m  <- αm m + βm
         3      1 2    3
    """
    comm = m1.dist.comm
    buf1 = m2.data
    xp = m1.xp

    N = m1.shape[0]
    n = (N + comm.size - 1) // comm.size

    for r in range(comm.size):
        if r == 0:
            buf2 = xp.empty((n, buf1.shape[1]), dtype=buf1.dtype)

        rrequest = None
        srequest = None
        if r < comm.size - 1:
            rrank = (comm.rank + r + 1) % comm.size
            rn1 = min(rrank * n, N)
            rn2 = min(rn1 + n, N)
            if rn2 > rn1:
                rrequest = comm.receive(buf2[:rn2 - rn1], rrank, 21, False)
            srank = (comm.rank - r - 1) % comm.size
            if len(m2.data) > 0:
                srequest = comm.send(m2.data, srank, 21, False)

        r0 = (comm.rank + r) % comm.size
        n1 = min(r0 * n, N)
        n2 = min(n1 + n, N)
        mmm(alpha, m1.data[:, n1:n2], 'N', buf1[:n2 - n1], 'N', beta, m3.data)

        beta = 1.0

        if r == 0:
            buf1 = xp.empty_like(buf2)

        buf1, buf2 = buf2, buf1

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

    return m3


def mmm_nc_sym(a, b, out, alpha, mmm):
    """Symmetric parallel matrix-matrix multiplication.

    :::

                †
        c <- αab + c

    Only lower half of c is updated.
    """
    comm = a.dist.comm
    M, N = a.shape
    m = (M + comm.size - 1) // comm.size
    mym = len(a.data)
    xp = a.xp

    buf1 = xp.empty((m, N), dtype=a.dtype)
    buf2 = xp.empty((m, N), dtype=a.dtype)
    half = comm.size // 2
    aa = a.data
    bb = b.data

    for r in range(half + 1):
        rrequest = None
        srequest = None

        if r < half:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            skip = (comm.size % 2 == 0 and r == half - 1)
            m1 = min(rrank * m, M)
            m2 = min(m1 + m, M)
            if not (skip and comm.rank < half) and m2 > m1:
                rrequest = comm.receive(buf1[:m2 - m1], rrank, 11, False)
            if not (skip and comm.rank >= half) and mym > 0:
                srequest = comm.send(b.data, srank, 11, False)

        if not (comm.size % 2 == 0 and r == half and comm.rank < half):
            m1 = min(((comm.rank - r) % comm.size) * m, M)
            m2 = min(m1 + m, M)
            if r == 0:
                # symmmmmmmmmmmmmmmmmmmmmmetricccccccccccccccc
                mmm(alpha, aa, 'N', bb, 'C', 1.0, out.data[:, m1:m2])
            else:
                beta = 1.0 if r <= comm.rank else 0.0
                mmm(alpha, aa, 'N', buf2[:m2 - m1], 'C',
                    beta, out.data[:, m1:m2])
            # out.data[:, m1:m2] = m12.data[:, :m2 - m1]

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        bb = buf1
        buf1, buf2 = buf2, buf1

    requests = []
    blocks = []
    nrows = (comm.size - 1) // 2
    for row in range(nrows):
        for column in range(comm.size - nrows + row, comm.size):
            if comm.rank == row:
                m1 = min(column * m, M)
                m2 = min(m1 + m, M)
                if mym > 0 and m2 > m1:
                    requests.append(
                        comm.send(out.data[:, m1:m2].T.conj().copy(),
                                  column, 12, False))
            elif comm.rank == column:
                m1 = min(row * m, M)
                m2 = min(m1 + m, M)
                if mym > 0 and m2 > m1:
                    block = xp.empty((mym, m2 - m1), out.dtype)
                    blocks.append((m1, m2, block))
                    requests.append(comm.receive(block, row, 12, False))

    comm.waitall(requests)
    for m1, m2, block in blocks:
        out.data[:, m1:m2] += block

    return out


def mmm_nc(a, b, out, alpha, beta, mmm):
    """Symmetric parallel matrix-matrix multiplication.

    :::

                †
        c <- αab  + βc
    """
    comm = a.dist.comm
    M, N = a.shape
    m = (M + comm.size - 1) // comm.size
    mym = len(a.data)
    xp = a.xp

    buf1 = xp.empty((m, N), dtype=a.dtype)
    buf2 = xp.empty((m, N), dtype=a.dtype)
    aa = a.data
    bb = b.data

    for r in range(comm.size):
        rrequest = None
        srequest = None

        if r < comm.size - 1:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            m1 = min(rrank * m, M)
            m2 = min(m1 + m, M)
            if m2 > m1:
                rrequest = comm.receive(buf1[:m2 - m1], rrank, 11, False)
            if mym > 0:
                srequest = comm.send(b.data, srank, 11, False)

        m1 = min(((comm.rank - r) % comm.size) * m, M)
        m2 = min(m1 + m, M)
        # symmmmmmmmmmmmmmmmmmmmmmetricccccccccccccccc ??
        mmm(alpha, aa, 'N', bb[:m2 - m1], 'C', beta, out.data[:, m1:m2])

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        bb = buf1
        buf1, buf2 = buf2, buf1

    return out



import pytest
import numpy as np
from ase import Atoms
from gpaw import GPAW, PW
from gpaw.response.bse import BSE


@pytest.mark.response
@pytest.mark.serial
def test_bse_spinpol(in_tmp_dir):
    """Make sure spinpolarized eigenvalues in gw_skn work.

    See issue #1066.
    """
    atoms = Atoms('H', magmoms=[1], pbc=True)
    atoms.center(vacuum=1.5)
    atoms.calc = GPAW(mode=PW(180, force_complex_dtype=True),
                      nbands=6,
                      convergence={'bands': 4})
    atoms.get_potential_energy()

    gw_skn = np.zeros((2, 1, 2))
    gw_skn[0, 0] = [-10, 2]
    gw_skn[1, 0] = [2, 4]

    bse = BSE(atoms.calc,
              ecut=10,
              nbands=2,
              gw_skn=gw_skn,
              valence_bands=[[0], [0]],
              conduction_bands=[[1], [1]])

    bsematrix = bse.get_bse_matrix()
    w_T, _ = bse.diagonalize_bse_matrix(bsematrix)
    assert w_T[0] == pytest.approx(0.013, abs=0.001)



from __future__ import annotations

from time import ctime
from typing import TYPE_CHECKING

import numpy as np
from ase.units import Ha

import gpaw
from gpaw.response import ResponseContext
from gpaw.response.chi0_data import (Chi0Data, Chi0BodyData,
                                     Chi0OpticalExtensionData)
from gpaw.response.frequencies import FrequencyDescriptor
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.hilbert import HilbertTransform
from gpaw.response import timer
from gpaw.response.pw_parallelization import PlaneWaveBlockDistributor
from gpaw.utilities.memory import maxrss
from gpaw.response.chi0_base import Chi0ComponentPWCalculator, Chi0Integrand
from gpaw.response.integrators import (
    HermitianOpticalLimit, HilbertOpticalLimit, OpticalLimit,
    HilbertOpticalLimitTetrahedron,
    Hermitian, Hilbert, HilbertTetrahedron, GenericUpdate)

if TYPE_CHECKING:
    from typing import Any
    from gpaw.typing import ArrayLike1D
    from gpaw.response.groundstate import ResponseGroundStateAdapter
    from gpaw.response.pair import ActualPairDensityCalculator


class Chi0Calculator:
    def __init__(self,
                 gs: ResponseGroundStateAdapter,
                 context: ResponseContext | None = None,
                 nblocks=1,
                 eshift=None,
                 intraband=True,
                 rate=0.0,
                 **kwargs):
        self.gs = gs
        self.context = context or ResponseContext()

        self.chi0_body_calc = Chi0BodyCalculator(
            self.gs, self.context,
            nblocks=nblocks, eshift=eshift, **kwargs)
        self.chi0_opt_ext_calc = Chi0OpticalExtensionCalculator(
            self.gs, self.context,
            intraband=intraband, rate=rate, **kwargs)

    @property
    def wd(self) -> FrequencyDescriptor:
        wd = self.chi0_body_calc.wd
        assert wd is self.chi0_opt_ext_calc.wd
        return wd

    @property
    def pair_calc(self) -> ActualPairDensityCalculator:
        # In a future refactor, we should find better ways to access the pair
        # density calculator (and the pair density paw corrections) XXX

        # pair_calc: ActualPairDensityCalculator from gpaw.response.pair
        return self.chi0_body_calc.pair_calc

    def create_chi0(self, q_c: list | np.ndarray) -> Chi0Data:

        # chi0_body: Chi0BodyData from gpaw.response.chi0_data
        chi0_body = self.chi0_body_calc.create_chi0_body(q_c)

        # chi0: Chi0Data from gpaw.response.chi0_data
        chi0 = Chi0Data.from_chi0_body_data(chi0_body)
        return chi0

    def calculate(self, q_c: list | np.ndarray) -> Chi0Data:
        """Calculate chi0 (possibly with optical extensions).

        Parameters
        ----------
        q_c : list or ndarray
            Momentum vector.

        Returns
        -------
        chi0 : Chi0Data
            Data object containing the chi0 data arrays along with basis
            representation descriptors and blocks distribution
        """
        # Calculate body

        # chi0_body: Chi0BodyData from gpaw.response.chi0_data
        chi0_body = self.chi0_body_calc.calculate(q_c)
        # SingleQPWDescriptor from gpaw.response.pair_functions
        qpd = chi0_body.qpd

        # Calculate optical extension
        if qpd.optical_limit:
            if self.chi0_body_calc.eshift is not None:
                raise NotImplementedError("No wings eshift available")
            chi0_opt_ext = self.chi0_opt_ext_calc.calculate(qpd=qpd)
        else:
            chi0_opt_ext = None

        self.context.print('\nFinished calculating chi0\n')

        return Chi0Data(chi0_body, chi0_opt_ext)

    @timer('Calculate CHI_0')
    def update_chi0(self,
                    chi0: Chi0Data,
                    m1: int, m2: int, spins: list) -> Chi0Data:
        """In-place calculation of chi0 (with optical extension).

        Parameters
        ----------
        chi0 : Chi0Data
            Data and representation object
        m1 : int
            Lower band cutoff for band summation
        m2 : int
            Upper band cutoff for band summation
        spins : list
            List of spin indices to include in the calculation

        Returns
        -------
        chi0 : Chi0Data
        """
        self.chi0_body_calc.update_chi0_body(chi0.body, m1, m2, spins)
        if chi0.optical_limit:
            if self.chi0_body_calc.eshift is not None:
                raise NotImplementedError("No wings eshift available")
            assert chi0.optical_extension is not None
            # Update the head and wings
            self.chi0_opt_ext_calc.update_chi0_optical_extension(
                chi0.optical_extension, m1, m2, spins)
        return chi0


class Chi0BodyCalculator(Chi0ComponentPWCalculator):
    def __init__(self, *args,
                 eshift: float | None = None,
                 **kwargs):
        """Construct the Chi0BodyCalculator.

        Parameters
        ----------
        eshift : float or None
            Energy shift of the conduction bands in eV.
        """
        self.eshift = eshift / Ha if eshift else eshift

        super().__init__(*args, **kwargs)

        if self.gs.metallic:
            assert self.eshift is None, \
                'A rigid energy shift cannot be applied to the conduction '\
                'bands if there is no band gap'

    def create_chi0_body(self, q_c: list | np.ndarray) -> Chi0BodyData:
        # qpd: SingleQPWDescriptor from gpaw.response.pair_functions
        qpd = self.get_pw_descriptor(q_c)
        return self._create_chi0_body(qpd)

    def _create_chi0_body(self, qpd: SingleQPWDescriptor) -> Chi0BodyData:
        return Chi0BodyData(self.wd, qpd, self.get_blockdist())

    def get_blockdist(self) -> PlaneWaveBlockDistributor:
        # integrator: Integrator from gpaw.response.integrators
        #    (or a child of this class)
        return PlaneWaveBlockDistributor(
            self.context.comm,  # _Communicator object from gpaw.mpi
            self.integrator.blockcomm,  # _Communicator object from gpaw.mpi
            self.integrator.kncomm)  # _Communicator object from gpaw.mpi

    def calculate(self, q_c: list | np.ndarray) -> Chi0BodyData:
        """Calculate the chi0 body.

        Parameters
        ----------
        q_c : list or ndarray
            Momentum vector.
        """
        # Construct the output data structure
        # qpd: SingleQPWDescriptor from gpaw.response.pair_functions
        qpd = self.get_pw_descriptor(q_c)
        self.print_info(qpd)
        # chi0_body: Chi0BodyData from gpaw.response.chi0_data
        chi0_body = self._create_chi0_body(qpd)

        # Integrate all transitions into partially filled and empty bands
        m1, m2 = self.get_band_transitions()
        self.update_chi0_body(chi0_body, m1, m2, spins=range(self.gs.nspins))

        return chi0_body

    def update_chi0_body(self,
                         chi0_body: Chi0BodyData,
                         m1: int, m2: int, spins: list | range):
        """In-place calculation of the body.

        Parameters
        ----------
        m1 : int
            Lower band cutoff for band summation
        m2 : int
            Upper band cutoff for band summation
        spins : list
            List of spin indices to include in the calculation
        """
        qpd = chi0_body.qpd

        # Reset PAW correction in case momentum has change
        pairden_paw_corr = self.gs.pair_density_paw_corrections
        self.pawcorr = pairden_paw_corr(chi0_body.qpd)

        self.context.print('Integrating chi0 body.')

        # domain: Domain from from gpaw.response.integrators
        # analyzer: PWSymmetryAnalyzer from gpaw.response.symmetry
        domain, analyzer, prefactor = self.get_integration_domain(qpd, spins)
        integrand = Chi0Integrand(self, qpd=qpd, analyzer=analyzer,
                                  optical=False, m1=m1, m2=m2)

        chi0_body.data_WgG[:] /= prefactor
        if self.hilbert:
            # Allocate a temporary array for the spectral function
            out_WgG = chi0_body.zeros()
        else:
            # Use the preallocated array for direct updates
            out_WgG = chi0_body.data_WgG
        self.integrator.integrate(domain=domain,  # Integration domain
                                  integrand=integrand,
                                  task=self.task,
                                  wd=self.wd,  # Frequency Descriptor
                                  out_wxx=out_WgG)  # Output array

        if self.hilbert:
            # The integrator only returns the spectral function and a Hilbert
            # transform is performed to return the real part of the density
            # response function.
            with self.context.timer('Hilbert transform'):
                # Make Hilbert transform
                ht = HilbertTransform(np.array(self.wd.omega_w), self.eta,
                                      timeordered=self.timeordered)
                ht(out_WgG)
            # Update the actual chi0 array
            chi0_body.data_WgG[:] += out_WgG
        chi0_body.data_WgG[:] *= prefactor

        tmp_chi0_wGG = chi0_body.copy_array_with_distribution('wGG')
        analyzer.symmetrize_wGG(tmp_chi0_wGG)
        chi0_body.data_WgG[:] = chi0_body.blockdist.distribute_as(
            tmp_chi0_wGG, chi0_body.nw, 'WgG')

    def construct_hermitian_task(self):
        return Hermitian(self.integrator.blockcomm, eshift=self.eshift)

    def construct_point_hilbert_task(self):
        return Hilbert(self.integrator.blockcomm, eshift=self.eshift)

    def construct_tetra_hilbert_task(self):
        return HilbertTetrahedron(self.integrator.blockcomm)

    def construct_literal_task(self):
        return GenericUpdate(
            self.eta, self.integrator.blockcomm, eshift=self.eshift)

    def print_info(self, qpd: SingleQPWDescriptor):

        if gpaw.dry_run:
            from gpaw.mpi import SerialCommunicator
            size = gpaw.dry_run
            comm = SerialCommunicator()
            comm.size = size
        else:
            comm = self.context.comm

        q_c = qpd.q_c
        nw = len(self.wd)
        csize = comm.size
        knsize = self.integrator.kncomm.size
        bsize = self.integrator.blockcomm.size
        chisize = nw * qpd.ngmax**2 * 16. / 1024**2 / bsize

        isl = ['', f'{ctime()}',
               'Calculating chi0 body with:',
               self.get_gs_info_string(tab='    '), '',
               '    Linear response parametrization:',
               f'    q_c: [{q_c[0]}, {q_c[1]}, {q_c[2]}]',
               self.get_response_info_string(qpd, tab='    '),
               f'    comm.size: {csize}',
               f'    kncomm.size: {knsize}',
               f'    blockcomm.size: {bsize}']
        if bsize > nw:
            isl.append(
                'WARNING! Your nblocks is larger than number of frequency'
                ' points. Errors might occur, if your submodule does'
                ' not know how to handle this.')
        isl.extend(['',
                    '    Memory estimate of potentially large arrays:',
                    f'        chi0_wGG: {chisize} M / cpu',
                    '        Memory usage before allocation: '
                    f'{(maxrss() / 1024**2)} M / cpu'])
        self.context.print('\n'.join(isl))


class Chi0OpticalExtensionCalculator(Chi0ComponentPWCalculator):

    def __init__(self, *args,
                 intraband=True,
                 rate=0.0,
                 **kwargs):
        """Contruct the Chi0OpticalExtensionCalculator.

        Parameters
        ----------
        intraband : bool
            Flag for including the intraband contribution to the chi0 head.
        rate : float, str
            Phenomenological scattering rate to use in optical limit Drude term
            (in eV). If rate='eta', it uses input artificial broadening eta as
            rate. Please note that for consistency the rate is implemented as
            omegap^2 / (omega + 1j * rate)^2, which differs from some
            literature by a factor of 2.
        """
        # Serial block distribution
        super().__init__(*args, nblocks=1, **kwargs)

        # In the optical limit of metals, one must add the Drude dielectric
        # response from the free-space plasma frequency of the intraband
        # transitions to the head of chi0. This is handled by a separate
        # calculator, provided that intraband is set to True.
        if self.gs.metallic and intraband:
            from gpaw.response.chi0_drude import Chi0DrudeCalculator
            if rate == 'eta':
                rate = self.eta * Ha  # external units
            self.rate = rate
            self.drude_calc = Chi0DrudeCalculator(
                self.gs, self.context,
                disable_point_group=self.disable_point_group,
                disable_time_reversal=self.disable_time_reversal,
                integrationmode=self.integrationmode)
        else:
            self.drude_calc = None
            self.rate = None

    def calculate(self,
                  qpd: SingleQPWDescriptor | None = None
                  ) -> Chi0OpticalExtensionData:
        """Calculate the chi0 head and wings."""
        # Create data object
        if qpd is None:
            qpd = self.get_pw_descriptor(q_c=[0., 0., 0.])

        # wd: FrequencyDescriptor from gpaw.response.frequencies
        chi0_opt_ext = Chi0OpticalExtensionData(self.wd, qpd)

        self.print_info(qpd)

        # Define band transitions
        m1, m2 = self.get_band_transitions()

        # Perform the actual integration
        self.update_chi0_optical_extension(chi0_opt_ext, m1, m2,
                                           spins=range(self.gs.nspins))

        if self.drude_calc is not None:
            # Add intraband contribution
            # drude_calc: Chi0DrudeCalculator from gpaw.response.chi0_drude
            # chi0_drude: Chi0DrudeData from gpaw.response.chi0_data
            chi0_drude = self.drude_calc.calculate(self.wd, self.rate)
            chi0_opt_ext.head_Wvv[:] += chi0_drude.chi_Zvv

        return chi0_opt_ext

    def update_chi0_optical_extension(
            self,
            chi0_optical_extension: Chi0OpticalExtensionData,
            m1: int, m2: int,
            spins: list | range):
        """In-place calculation of the chi0 head and wings.

        Parameters
        ----------
        m1 : int
            Lower band cutoff for band summation
        m2 : int
            Upper band cutoff for band summation
        spins : list
            List of spin indices to include in the calculation
        """
        self.context.print('Integrating chi0 head and wings.')
        chi0_opt_ext = chi0_optical_extension
        qpd = chi0_opt_ext.qpd

        domain, analyzer, prefactor = self.get_integration_domain(qpd, spins)
        integrand = Chi0Integrand(self, qpd=qpd, analyzer=analyzer,
                                  optical=True, m1=m1, m2=m2)

        # We integrate the head and wings together, using the combined index P
        # index v = (x, y, z)
        # index G = (G0, G1, G2, ...)
        # index P = (x, y, z, G1, G2, ...)
        WxvP_shape = list(chi0_opt_ext.WxvG_shape)
        WxvP_shape[-1] += 2
        tmp_chi0_WxvP = np.zeros(WxvP_shape, complex)
        self.integrator.integrate(domain=domain,  # Integration domain
                                  integrand=integrand,
                                  task=self.task,
                                  wd=self.wd,  # Frequency Descriptor
                                  out_wxx=tmp_chi0_WxvP)  # Output array
        if self.hilbert:
            with self.context.timer('Hilbert transform'):
                ht = HilbertTransform(np.array(self.wd.omega_w), self.eta,
                                      timeordered=self.timeordered)
                ht(tmp_chi0_WxvP)
        tmp_chi0_WxvP *= prefactor

        # Fill in wings part of the data, but leave out the head part (G0)
        chi0_opt_ext.wings_WxvG[..., 1:] += tmp_chi0_WxvP[..., 3:]
        analyzer.symmetrize_wxvG(chi0_opt_ext.wings_WxvG)
        # Fill in the head
        chi0_opt_ext.head_Wvv[:] += tmp_chi0_WxvP[:, 0, :3, :3]
        analyzer.symmetrize_wvv(chi0_opt_ext.head_Wvv)

    def construct_hermitian_task(self):
        return HermitianOpticalLimit()

    def construct_point_hilbert_task(self):
        return HilbertOpticalLimit()

    def construct_tetra_hilbert_task(self):
        return HilbertOpticalLimitTetrahedron()

    def construct_literal_task(self):
        return OpticalLimit(eta=self.eta)

    def print_info(self, qpd: SingleQPWDescriptor):
        """Print information about optical extension calculation."""
        isl = ['',
               f'{ctime()}',
               'Calculating chi0 optical extensions with:',
               self.get_gs_info_string(tab='    '),
               '',
               '    Linear response parametrization:',
               self.get_response_info_string(qpd, tab='    ')]
        self.context.print('\n'.join(isl))


def get_frequency_descriptor(
        frequencies: ArrayLike1D | dict[str, Any] | None = None, *,
        gs: ResponseGroundStateAdapter | None = None,
        nbands: int | None = None):
    """Helper function to generate frequency descriptors.

    In most cases, the `frequencies` input can be processed directly via
    wd = FrequencyDescriptor.from_array_or_dict(frequencies),
    but in cases where `frequencies` does not specify omegamax, it is
    calculated from the input ground state adapter.
    """
    if frequencies is None:
        frequencies = {'type': 'nonlinear'}  # default frequency grid
    if isinstance(frequencies, dict) and frequencies.get('omegamax') is None:
        assert gs is not None
        frequencies['omegamax'] = get_omegamax(gs, nbands)
    return FrequencyDescriptor.from_array_or_dict(frequencies)


def get_omegamax(gs: ResponseGroundStateAdapter,
                 nbands: int | None = None):
    """Get the maxmimum eigenvalue difference including nbands, in eV."""
    epsmin, epsmax = gs.get_eigenvalue_range(nbands=nbands)
    return (epsmax - epsmin) * Ha



from ase.build import molecule
from gpaw import GPAW, PW
from gpaw.response.g0w0 import G0W0
from gpaw.mpi import world
import pytest


@pytest.mark.response
def test_gw_sg15(in_tmp_dir, add_cwd_to_setup_paths):
    from gpaw.test.pseudopotential.H_sg15 import pp_text
    if world.rank == 0:
        with open('H_ONCV_PBE-1.0.upf', 'w') as pp_file:
            print(pp_text, file=pp_file)
    world.barrier()
    sys = molecule('H2', pbc=True)
    sys.center(vacuum=2.5)

    calc = GPAW(setups='sg15',
                xc='PBE', mode=PW(ecut=300), convergence={'bands': 40},
                nbands=50, kpts=(2, 2, 2))
    sys.calc = calc
    sys.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    gw = G0W0(calc='gs.gpw',
              bands=(1, 6),
              ecut=20,
              nblocksmax=True,
              filename='H2_g0w0_b11-15')
    results = gw.calculate()
    refs = pytest.approx([1.58177132, 3.33595453, 3.87312694, 4.95358262,
                         6.09555254], abs=1e-5)
    assert results['qp'][0][0] == refs



import numpy as np
import pytest
from ase.build import molecule
from ase.parallel import parprint, world

from gpaw import GPAW
from gpaw.analyse.overlap import Overlap
from gpaw.utilities.adjust_cell import adjust_cell
from gpaw.lrtddft import LrTDDFT

"""Evaluate the overlap between two independent calculations

Differences are forced by different eigensolvers and differing number
of Kohn-Sham states.
"""


@pytest.mark.later
def test_overlap(in_tmp_dir):

    def get_kwargs(nbands, **kwargs) -> dict:
        base_kwargs = dict(
            mode='fd',
            h=h,
            txt=txt,
            nbands=nbands,
            convergence={'eigenstates': nbands})
        return dict(base_kwargs, **kwargs)

    h = 0.4
    box = 2
    nbands = 4
    txt = '-'
    txt = None

    H2 = molecule('H2')
    adjust_cell(H2, box, h)

    c1 = GPAW(**get_kwargs(eigensolver='dav', nbands=nbands))
    c1.calculate(H2)
    lr1 = LrTDDFT(c1)

    parprint('sanity --------')
    ov = Overlap(c1).pseudo(c1)
    parprint('pseudo(normalized):\n', ov)
    ov = Overlap(c1).pseudo(c1, False)
    parprint('pseudo(not normalized):\n', ov)
    ov = Overlap(c1).full(c1)
    parprint('full:\n', ov)
    assert ov[0] == pytest.approx(np.eye(ov[0].shape[0], dtype=ov.dtype),
                                  abs=1e-10)

    def show(c2):
        c2.calculate(H2)
        ov = Overlap(c1).pseudo(c2)
        parprint('wave function overlap (pseudo):\n', ov)
        ov = Overlap(c1).full(c2)
        parprint('wave function overlap (full):\n', ov)
        lr2 = LrTDDFT(c2)
        ovkss = lr1.kss.overlap(ov[0], lr2.kss)
        parprint('KSSingles overlap:\n', ovkss)
        ovlr = lr1.overlap(ov[0], lr2)
        parprint('LrTDDFT overlap:\n', ovlr)

    parprint('cg --------')
    c2 = GPAW(**get_kwargs(eigensolver='cg', nbands=nbands + 1))
    show(c2)

    parprint('spin --------')
    H2.set_initial_magnetic_moments([1, -1])
    c2 = GPAW(
        **get_kwargs(
            spinpol=True, nbands=nbands + 1, parallel={'domain': world.size}))
    H2.set_initial_magnetic_moments([0, 0])
    try:
        show(c2)
    except AssertionError:
        parprint('Not ready')
    if 1:
        ov = Overlap(c1).pseudo(c2, otherspin=1)
        parprint('wave function overlap (pseudo other spin):\n', ov)

    parprint('k-points --------')

    H2.set_pbc([1, 1, 1])
    c1 = GPAW(**get_kwargs(nbands=nbands, kpts=(1, 1, 3)))
    c1.calculate(H2)
    c2 = GPAW(**get_kwargs(nbands=nbands + 1, kpts=(1, 1, 3)))
    try:
        show(c2)
    except (AssertionError, IndexError) as e:
        parprint('Not ready', e)



import pytest
import numpy as np

from gpaw.core import UGDesc
from gpaw.gpu import cupy as cp
from gpaw.mpi import world
from gpaw.spline import Spline
from gpaw.new.c import GPU_AWARE_MPI


@pytest.mark.parametrize('dtype', [float, complex])
def test_acf(dtype):
    if world.size > 1 and not GPU_AWARE_MPI:
        pytest.skip('MPI not GPU-aware')
    s = Spline.from_data(0, 1.0, [1.0, 0.5, 0.0])
    n = 40
    a = 8.0
    fracpos_ac = [(0.5, 0.5, 0.25 + 0.25 * i) for i in [0, 1, 2]]

    grid = UGDesc(cell=[a, a, a], size=(n, n, n), comm=world, dtype=dtype)
    basis_cpu = grid.atom_centered_functions([[s], [s], [s]],
                                             positions=fracpos_ac, xp=np)
    basis_gpu = grid.atom_centered_functions([[s], [s], [s]],
                                             positions=fracpos_ac, xp=cp)

    P_cpu_ani = basis_cpu.layout.empty()
    P_gpu_ani = basis_gpu.layout.empty()
    P_cpu_ani.data[:] = 1.0
    P_gpu_ani.data[:] = 1.0

    b_cpu = grid.zeros(xp=np)
    b_gpu = grid.zeros(xp=cp)
    basis_cpu.add_to(b_cpu, P_cpu_ani)
    basis_gpu.add_to(b_gpu, P_gpu_ani)

    assert b_cpu.data == pytest.approx(b_gpu.data.get(), abs=1e-12)

    out_cpu_ani = basis_cpu.integrate(b_cpu)
    out_gpu_ani = basis_gpu.integrate(b_gpu)

    for a, out_cpu_ni in out_cpu_ani.items():
        assert out_cpu_ni == pytest.approx(out_gpu_ani[a].get(), abs=1e-12)



import pytest
from ase.build import bulk
from gpaw import GPAW
from gpaw.poisson import FFTPoissonSolver, FDPoissonSolver, FastPoissonSolver
import numpy as np


@pytest.mark.parametrize('poisson', [FFTPoissonSolver, FDPoissonSolver,
                                     FastPoissonSolver])
def test_poisson_cell_change(poisson):
    atoms = bulk('Na')
    calc = GPAW(mode='lcao', poissonsolver=poisson())
    atoms.calc = calc
    atoms.get_potential_energy()
    atoms.set_cell(atoms.cell * 1.1, scale_atoms=True)
    Ep = atoms.get_potential_energy()
    calc = GPAW(mode='lcao', poissonsolver=poisson())
    atoms.calc = calc
    Ep2 = atoms.get_potential_energy()
    assert np.allclose(Ep, Ep2, rtol=1e-4, atol=1e-5)



import pytest


def test_gap_in_txt(gpw_files, needs_ase_master):
    gpw = gpw_files['h2_pw']
    txt = gpw.with_suffix('.txt')
    for line in txt.read_text().splitlines():
        if line.startswith('Gap:'):
            gap = float(line.split()[1])
            assert gap == pytest.approx(11.296, abs=0.001)
            return
    raise ValueError(f'No gap in text file: {txt}')



import warnings
import numpy as np

from ase.units import Ha

from gpaw.directmin.tools import (sort_orbitals_according_to_energies,
                                  get_n_occ)


def do_if_converged(eigensolver_name, wfs, ham, dens, log):
    name = eigensolver_name
    if name == 'etdm-lcao' or name == 'etdm-fdpw':
        occ_name = getattr(wfs.occupations, 'name', None)
        solver = wfs.eigensolver
        if hasattr(solver, 'dm_helper'):
            func_name = solver.dm_helper.func.name
        elif hasattr(solver, 'odd'):
            func_name = solver.odd.name
        sic_calc = 'SIC' in func_name
    else:
        return

    if hasattr(solver, 'e_sic'):
        e_sic = solver.e_sic
    else:
        e_sic = 0.0

    if eigensolver_name == 'etdm-lcao':
        with ((wfs.timer('Get canonical representation'))):
            for kpt in wfs.kpt_u:
                solver.dm_helper.update_to_canonical_orbitals(
                    wfs, ham, kpt, False, False)

        if sic_calc:
            if hasattr(solver, 'constraints'):
                constraints = solver.constraints
            else:
                constraints = None
            # Sort orbitals according to orbital energies
            sort_orbitals_according_to_energies(
                ham, wfs, constraints, use_eps=False)

        solver.set_ref_orbitals_and_a_vec(wfs)

        log('\nOccupied states converged after'
            ' {:d} e/g evaluations'.format(solver.eg_count))

    elif eigensolver_name == 'etdm-fdpw':
        solver.choose_optimal_orbitals(wfs)
        niter1 = solver.eg_count
        niter2 = 0
        niter3 = 0

        iloop1 = solver.iloop is not None
        iloop2 = solver.outer_iloop is not None
        if iloop1:
            niter2 = solver.total_eg_count_iloop
        if iloop2:
            niter3 = solver.total_eg_count_outer_iloop

        if iloop1 and iloop2:
            log(
                '\nOccupied states converged after'
                ' {:d} KS and {:d} SIC e/g '
                'evaluations'.format(niter3,
                                     niter2 + niter3))
        elif not iloop1 and iloop2:
            log(
                '\nOccupied states converged after'
                ' {:d} e/g evaluations'.format(niter3))
        elif iloop1 and not iloop2:
            log(
                '\nOccupied states converged after'
                ' {:d} KS and {:d} SIC e/g '
                'evaluations'.format(niter1, niter2))
        else:
            log(
                '\nOccupied states converged after'
                ' {:d} e/g evaluations'.format(niter1))
        if solver.converge_unocc:
            log('Converge unoccupied states:')
            max_er = wfs.eigensolver.error
            max_er *= Ha ** 2 / wfs.nvalence
            solver.run_unocc(ham, wfs, dens, max_er, log)
        else:
            log('Unoccupied states are not converged.')
        solver.initialized = False

        rewrite_psi = True
        if sic_calc:
            rewrite_psi = False

        solver.get_canonical_representation(ham, wfs, rewrite_psi)

    if occ_name == 'mom':
        check_mom_no_update_of_occupations(wfs)

    solver.update_ks_energy(ham, wfs, dens)
    ham.get_energy(0.0, wfs, kin_en_using_band=False, e_sic=e_sic)

    if occ_name == 'mom' and not sic_calc:
        if hasattr(solver, 'constraints'):
            constraints = solver.constraints
        else:
            constraints = None
        # Sort orbitals according to eigenvalues
        sort_orbitals_according_to_energies(
            ham, wfs, constraints, use_eps=True)
        not_update = not wfs.occupations.update_numbers
        fixed_occ = wfs.occupations.use_fixed_occupations
        if not_update or fixed_occ:
            wfs.occupations.numbers = solver.initial_occupation_numbers


def check_eigensolver_state(eigensolver_name, wfs, ham, dens, log):
    solver = wfs.eigensolver
    name = eigensolver_name
    if name == 'etdm-lcao' or name == 'etdm-fdpw':
        solver.eg_count = 0
        solver.globaliters = 0

        if hasattr(solver, 'iloop'):
            if solver.iloop is not None:
                solver.iloop.total_eg_count = 0
        if hasattr(solver, 'outer_iloop'):
            if solver.outer_iloop is not None:
                solver.outer_iloop.total_eg_count = 0

        solver.check_assertions(wfs, dens)
        if (hasattr(solver, 'dm_helper') and solver.dm_helper is None) \
                or not solver.initialized:
            solver.initialize_dm_helper(wfs, ham, dens, log)


def check_mom_no_update_of_occupations(wfs):
    f_sn = wfs.occupations.update_occupations()
    for kpt in wfs.kpt_u:
        k = wfs.kd.nibzkpts * kpt.s + kpt.q
        n_occ, occupied = get_n_occ(kpt)
        if n_occ != 0.0 and np.min(f_sn[k][:n_occ]) == 0:
            warnings.warn('MOM has detected variational collapse '
                          'after getting canonical orbitals. Check '
                          'that the orbitals are consistent with the '
                          'initial guess.')



from dataclasses import dataclass
from datetime import timedelta
from functools import cached_property
from time import time, ctime

from ase.units import Hartree
from ase.dft import monkhorst_pack
import numpy as np
from scipy.linalg import eigh

from gpaw.blacs import BlacsGrid, Redistributor
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.mpi import world, serial_comm
from gpaw.response import ResponseContext
from gpaw.response.groundstate import CellDescriptor
from gpaw.response.chi0 import Chi0Calculator
from gpaw.response.context import timer
from gpaw.response.coulomb_kernels import CoulombKernel
from gpaw.response.df import write_response_function
from gpaw.response.frequencies import FrequencyDescriptor
from gpaw.response.pair import KPointPairFactory, get_gs_and_context
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.screened_interaction import initialize_w_calculator


def decide_whether_tammdancoff(val_sn, con_sn):
    for n in val_sn[0]:
        if n in con_sn[0]:
            return False
    if len(val_sn) == 2:
        for n in val_sn[1]:
            if n in con_sn[1]:
                return False
    return True


@dataclass
class BSEMatrix:
    rhoG0_S: np.ndarray
    df_S: np.ndarray
    H_sS: np.ndarray

    def diagonalize_nontamdancoff(self, bse):
        df_S = self.df_S
        H_sS = self.H_sS
        rhoG0_S = self.rhoG0_S

        excludef_S = np.where(np.abs(df_S) < 0.001)[0]
        bse.context.print('  Using numpy.linalg.eig...')
        bse.context.print('  Eliminated %s pair orbitals' % len(
            excludef_S))

        H_SS = bse.collect_A_SS(H_sS)
        w_T = np.zeros(bse.nS - len(excludef_S), complex)
        if world.rank == 0:
            H_SS = np.delete(H_SS, excludef_S, axis=0)
            H_SS = np.delete(H_SS, excludef_S, axis=1)
            w_T, v_ST = np.linalg.eig(H_SS)
        world.broadcast(w_T, 0)

        # Here the eigenvectors are represented as complex conjugated rows

        df_S = np.delete(df_S, excludef_S)
        rhoG0_S = np.delete(rhoG0_S, excludef_S)

        C_T = np.zeros(bse.nS - len(excludef_S), complex)
        if world.rank == 0:
            A_T = np.dot(rhoG0_S, v_ST)
            B_T = np.dot(rhoG0_S * df_S, v_ST)
            tmp = np.dot(v_ST.conj().T, v_ST)
            overlap_tt = np.linalg.inv(tmp)
            C_T = np.dot(B_T.conj(), overlap_tt.T) * A_T
        world.broadcast(C_T, 0)
        return w_T, C_T

    def diagonalize_tamdancoff(self, bse):
        df_S = self.df_S
        H_sS = self.H_sS
        rhoG0_S = self.rhoG0_S

        nS = bse.nS
        ns = bse.ns

        if world.size == 1:
            bse.context.print('  Using lapack...')
            w_T, v_St = eigh(H_sS)
        else:
            bse.context.print('  Using scalapack...')
            assert ns == (
                -(-bse.kd.nbzkpts // world.size) * (
                    bse.nv * bse.nc *
                    bse.spins *
                    (bse.spinors + 1)**2))

            # XXX We don't need to create new BLACS grids all the time
            # (also: remove the one further down)
            grid = BlacsGrid(world, world.size, 1)
            desc = grid.new_descriptor(nS, nS, ns, nS)

            desc2 = grid.new_descriptor(nS, nS, 2, 2)
            H_tmp = desc2.zeros(dtype=complex)
            r = Redistributor(world, desc, desc2)
            r.redistribute(H_sS, H_tmp)

            w_T = np.empty(nS)
            v_tmp = desc2.empty(dtype=complex)
            desc2.diagonalize_dc(H_tmp, v_tmp, w_T)

            r = Redistributor(grid.comm, desc2, desc)
            v_St = desc.zeros(dtype=complex)
            r.redistribute(v_tmp, v_St)
            v_St = v_St.conj().T

        A_t = np.dot(rhoG0_S, v_St)
        B_t = np.dot(rhoG0_S * df_S, v_St)

        if world.size == 1:
            C_T = B_t.conj() * A_t
        else:
            grid = BlacsGrid(world, world.size, 1)
            desc = grid.new_descriptor(nS, 1, ns, 1)
            C_t = desc.empty(dtype=complex)
            C_t[:, 0] = B_t.conj() * A_t
            C_T = desc.collect_on_master(C_t)[:, 0]
            if world.rank != 0:
                C_T = np.empty(nS, dtype=complex)
            world.broadcast(C_T, 0)
        return w_T, C_T


@dataclass
class ScreenedPotential:
    pawcorr_q: list
    W_qGG: list
    qpd_q: list


class BSEBackend:
    def __init__(self, *, gs, context,
                 valence_bands, conduction_bands,
                 spinors=False,
                 ecut=10.,
                 scale=1.0,
                 nbands=None,
                 eshift=None,
                 gw_skn=None,
                 truncation=None,
                 integrate_gamma=1,
                 mode='BSE',
                 q_c=[0.0, 0.0, 0.0],
                 direction=0):
        self.gs = gs
        self.q_c = q_c
        self.direction = direction
        self.context = context

        self.spinors = spinors
        self.scale = scale

        assert mode in ['RPA', 'BSE']

        self.ecut = ecut / Hartree
        self.nbands = nbands
        self.mode = mode

        if integrate_gamma == 0 and truncation is not None:
            self.context.print('***WARNING*** Analytical Coulomb integration' +
                               ' is not expected to work with Coulomb ' +
                               'truncation. Use integrate_gamma=1')
        self.integrate_gamma = integrate_gamma

        # Find q-vectors and weights in the IBZ:
        self.kd = self.gs.kd
        if -1 in self.kd.bz2bz_ks:
            self.context.print('***WARNING*** Symmetries may not be right. '
                               'Use gamma-centered grid to be sure')
        offset_c = 0.5 * ((self.kd.N_c + 1) % 2) / self.kd.N_c
        bzq_qc = monkhorst_pack(self.kd.N_c) + offset_c
        self.qd = KPointDescriptor(bzq_qc)
        self.qd.set_symmetry(self.gs.atoms, self.kd.symmetry)

        # bands
        self.spins = self.gs.nspins
        if self.spins == 2:
            if self.spinors:
                self.spinors = False
                self.context.print('***WARNING*** Presently the spinor ' +
                                   'version does not work for spin-polarized' +
                                   ' calculations. Performing scalar ' +
                                   'calculation')

        self.val_sn = self.parse_bands(valence_bands, band_type='valence')
        self.con_sn = self.parse_bands(conduction_bands,
                                       band_type='conduction')

        self.use_tammdancoff = decide_whether_tammdancoff(self.val_sn,
                                                          self.con_sn)

        self.nv = len(self.val_sn[0])
        self.nc = len(self.con_sn[0])
        if eshift is not None:
            eshift /= Hartree
        if gw_skn is not None:
            assert self.nv + self.nc == len(gw_skn[0, 0])
            assert self.kd.nibzkpts == len(gw_skn[0])
            gw_skn = gw_skn[:, self.kd.bz2ibz_k]
            # assert self.kd.nbzkpts == len(gw_skn[0])
            gw_skn /= Hartree
        self.gw_skn = gw_skn
        self.eshift = eshift

        # Number of pair orbitals
        self.nS = self.kd.nbzkpts * self.nv * self.nc * self.spins
        self.nS *= (self.spinors + 1)**2

        self.coulomb = CoulombKernel.from_gs(self.gs, truncation=truncation)
        self.context.print(self.coulomb.description())

        self.print_initialization(self.use_tammdancoff, self.eshift,
                                  self.gw_skn)

        self.Nv = self.nv * (self.spinors + 1)
        self.Nc = self.nc * (self.spinors + 1)
        self.ns = (
            -(-self.kd.nbzkpts // world.size) * self.Nv * self.Nc * self.spins)

    def parse_bands(self, bands, band_type='valence'):
        """Helper function that checks whether bands are correctly specified,
         and brings them to the format used later in the code.

        If the calculation is spin-polarized, band indices must
        be given explicitly as lists/arrays of shape (2,nbands) where the first
        index is for spin.

        If the calculation is not spin-polarized, either an integer (number of
        desired bands) or lists of band indices must be provided.

        band_type is an optional parameter that is only when a desired number
        of bands is given (rather than a list) to help figure out the correct
        band indices.
        """
        if hasattr(bands, '__iter__'):
            if self.spins == 2:
                if len(bands) != 2 or (len(bands[0]) != len(bands[1])):
                    raise ValueError('For a spin-polarized calculation, '
                                     'the same number of bands must be '
                                     'specified for each spin! valence and '
                                     'conduction bands must be lists of shape '
                                     '(2,n)')

            bands_sn = np.atleast_2d(bands)
            return bands_sn

        # if we get here, bands is not iterable
        # check that the specified input is valid

        if self.spins == 2:
            raise NotImplementedError('For a spin-polarized calculation, '
                                      'bands must be specified as lists '
                                      'of shape (2,n)')

        n_fully_occupied_bands, n_partially_occupied_bands = \
            self.gs.count_occupied_bands()

        if n_fully_occupied_bands != n_partially_occupied_bands:
            raise NotImplementedError('Automatic band generation is currently '
                                      'not implemented for metallic systems. '
                                      'Please specify band indices manually.')

        if band_type == 'valence':
            bands_sn = range(n_fully_occupied_bands - bands,
                             n_fully_occupied_bands)
        elif band_type == 'conduction':
            bands_sn = range(n_fully_occupied_bands,
                             n_fully_occupied_bands + bands)
        else:
            raise ValueError(f'Invalid band type: {band_type}')

        bands_sn = np.atleast_2d(bands_sn)
        return bands_sn

    @timer('BSE calculate')
    def calculate(self, optical):
        if self.spinors:
            # Calculate spinors. Here m is index of eigenvalues with SOC
            # and n is the basis of eigenstates without SOC. Below m is used
            # for unoccupied states and n is used for occupied states so be
            # careful!

            self.context.print('Diagonalizing spin-orbit Hamiltonian')
            soc = self.gs.soc_eigenstates(scale=self.scale)
            e_mk = soc.eigenvalues().T
            v_kmn = soc.eigenvectors()
            e_mk /= Hartree

        # Parallelization stuff
        nK = self.kd.nbzkpts
        myKrange, myKsize, mySsize = self.parallelisation_sizes()

        # Calculate exchange interaction
        qpd0 = SingleQPWDescriptor.from_q(self.q_c, self.ecut, self.gs.gd)
        ikq_k = self.kd.find_k_plus_q(self.q_c)
        v_G = self.coulomb.V(qpd=qpd0, q_v=None)

        if optical:
            v_G[0] = 0.0

        kptpair_factory = KPointPairFactory(
            gs=self.gs,
            context=ResponseContext(txt='pair.txt', timer=self.context.timer,
                                    comm=serial_comm))
        pair_calc = kptpair_factory.pair_calculator()
        pawcorr = self.gs.pair_density_paw_corrections(qpd0)

        if self.mode != 'RPA':
            screened_potential = self.calculate_screened_potential()

        # Calculate pair densities, eigenvalues and occupations
        self.context.timer.start('Pair densities')
        so = self.spinors + 1
        Nv = self.Nv
        Nc = self.Nc
        Ns = self.spins
        rhoex_KsmnG = np.zeros((nK, Ns, Nv, Nc, len(v_G)), complex)
        df_Ksmn = np.zeros((nK, Ns, Nv, Nc), float)  # -(ev - ec)
        deps_ksmn = np.zeros((myKsize, Ns, Nv, Nc), float)  # -(fv - fc)

        optical_limit = np.allclose(self.q_c, 0.0)

        get_pair = kptpair_factory.get_kpoint_pair
        get_pair_density = pair_calc.get_pair_density
        if self.spinors:
            # Get all pair densities to allow for SOC mixing
            # Use twice as many no-SOC states as BSE bands to allow mixing
            vi_s = [2 * self.val_sn[0, 0] - self.val_sn[0, -1] - 1]
            vf_s = [2 * self.con_sn[0, -1] - self.con_sn[0, 0] + 2]
            if vi_s[0] < 0:
                vi_s[0] = 0
            ci_s, cf_s = vi_s, vf_s
            ni, nf = vi_s[0], vf_s[0]
            mvi = 2 * self.val_sn[0, 0]
            mvf = 2 * (self.val_sn[0, -1] + 1)
            mci = 2 * self.con_sn[0, 0]
            mcf = 2 * (self.con_sn[0, -1] + 1)
        else:
            vi_s, vf_s = self.val_sn[:, 0], self.val_sn[:, -1] + 1
            ci_s, cf_s = self.con_sn[:, 0], self.con_sn[:, -1] + 1
        for ik, iK in enumerate(myKrange):
            for s in range(Ns):
                pair = get_pair(qpd0, s, iK,
                                vi_s[s], vf_s[s], ci_s[s], cf_s[s])
                m_m = np.arange(vi_s[s], vf_s[s])
                n_n = np.arange(ci_s[s], cf_s[s])
                if self.gw_skn is not None:
                    iKq = self.gs.kd.find_k_plus_q(self.q_c, [iK])[0]
                    epsv_m = self.gw_skn[s, iK, :self.nv]
                    epsc_n = self.gw_skn[s, iKq, self.nv:]
                    deps_ksmn[ik, s] = -(epsv_m[:, np.newaxis] - epsc_n)
                elif self.spinors:
                    iKq = self.gs.kd.find_k_plus_q(self.q_c, [iK])[0]
                    epsv_m = e_mk[mvi:mvf, iK]
                    epsc_n = e_mk[mci:mcf, iKq]
                    deps_ksmn[ik, s] = -(epsv_m[:, np.newaxis] - epsc_n)
                else:
                    deps_ksmn[ik, s] = -pair.get_transition_energies(m_m, n_n)

                rho_mnG = get_pair_density(qpd0, pair, m_m, n_n,
                                           pawcorr=pawcorr)
                if optical_limit:
                    n_mnv = pair_calc.get_optical_pair_density_head(
                        qpd0, pair, m_m, n_n)
                    rho_mnG[:, :, 0] = n_mnv[:, :, self.direction]
                if self.spinors:
                    v0_kmn = v_kmn[:, :, ::2]
                    v1_kmn = v_kmn[:, :, 1::2]
                    if optical_limit:
                        deps0_mn = -pair.get_transition_energies(m_m, n_n)
                        rho_mnG[:, :, 0] *= deps0_mn

                    # This recreates the old behaviour of
                    # get_occupation_differences(self.val_sn[s],self.con_sn[s])
                    df_mn = (pair.kpt1.f_n[self.val_sn[s] -
                                           pair.kpt1.n1][:, np.newaxis] -
                             pair.kpt2.f_n[self.con_sn[s] - pair.kpt2.n1])

                    df_Ksmn[iK, s, ::2, ::2] = df_mn
                    df_Ksmn[iK, s, ::2, 1::2] = df_mn
                    df_Ksmn[iK, s, 1::2, ::2] = df_mn
                    df_Ksmn[iK, s, 1::2, 1::2] = df_mn
                    vecv0_mn = v0_kmn[iK, mvi:mvf, ni:nf]
                    vecc0_mn = v0_kmn[iKq, mci:mcf, ni:nf]
                    rho_0mnG = np.dot(vecv0_mn.conj(),
                                      np.dot(vecc0_mn, rho_mnG))
                    vecv1_mn = v1_kmn[iK, mvi:mvf, ni:nf]
                    vecc1_mn = v1_kmn[iKq, mci:mcf, ni:nf]
                    rho_1mnG = np.dot(vecv1_mn.conj(),
                                      np.dot(vecc1_mn, rho_mnG))
                    rhoex_KsmnG[iK, s] = rho_0mnG + rho_1mnG
                    if optical_limit:
                        rhoex_KsmnG[iK, s, :, :, 0] /= deps_ksmn[ik, s]
                else:
                    df_Ksmn[iK, s] = pair.get_occupation_differences()
                    rhoex_KsmnG[iK, s] = rho_mnG

        if self.eshift is not None:
            deps_ksmn[np.where(df_Ksmn[myKrange] > 1.0e-3)] += self.eshift
            deps_ksmn[np.where(df_Ksmn[myKrange] < -1.0e-3)] -= self.eshift

        world.sum(df_Ksmn)
        world.sum(rhoex_KsmnG)

        rhoG0_S = np.reshape(rhoex_KsmnG[:, :, :, :, 0], -1)
        self.context.timer.stop('Pair densities')

        # Calculate Hamiltonian
        self.context.timer.start('Calculate Hamiltonian')
        t0 = time()
        self.context.print('Calculating {} matrix elements at q_c = {}'.format(
            self.mode, self.q_c))
        H_ksmnKsmn = np.zeros((myKsize, Ns, Nv, Nc, nK, Ns, Nv, Nc), complex)
        for ik1, iK1 in enumerate(myKrange):
            for s1 in range(Ns):
                kptv1 = kptpair_factory.get_k_point(
                    s1, iK1, vi_s[s1], vf_s[s1])
                kptc1 = kptpair_factory.get_k_point(
                    s1, ikq_k[iK1], ci_s[s1], cf_s[s1])
                rho1_mnG = rhoex_KsmnG[iK1, s1]

                rho1ccV_mnG = rho1_mnG.conj()[:, :] * v_G
                for s2 in range(Ns):
                    for Q_c in self.qd.bzk_kc:
                        iK2 = self.kd.find_k_plus_q(Q_c, [kptv1.K])[0]
                        rho2_mnG = rhoex_KsmnG[iK2, s2]
                        self.context.timer.start('Coulomb')
                        H_ksmnKsmn[ik1, s1, :, :, iK2, s2, :, :] += np.einsum(
                            'ijk,mnk->ijmn', rho1ccV_mnG, rho2_mnG,
                            optimize='optimal')
                        self.context.timer.stop('Coulomb')

                        if self.mode != 'RPA' and s1 == s2:
                            ikq = ikq_k[iK2]
                            kptv2 = kptpair_factory.get_k_point(
                                s1, iK2, vi_s[s1], vf_s[s1])
                            kptc2 = kptpair_factory.get_k_point(
                                s1, ikq, ci_s[s1], cf_s[s1])
                            rho3_mmG, iq = self.get_density_matrix(
                                pair_calc, screened_potential, kptv1, kptv2)
                            rho4_nnG, iq = self.get_density_matrix(
                                pair_calc, screened_potential, kptc1, kptc2)
                            if self.spinors:
                                vec0_mn = v0_kmn[iK1, mvi:mvf, ni:nf]
                                vec1_mn = v1_kmn[iK1, mvi:mvf, ni:nf]
                                vec2_mn = v0_kmn[iK2, mvi:mvf, ni:nf]
                                vec3_mn = v1_kmn[iK2, mvi:mvf, ni:nf]
                                rho_0mnG = np.dot(vec0_mn.conj(),
                                                  np.dot(vec2_mn, rho3_mmG))
                                rho_1mnG = np.dot(vec1_mn.conj(),
                                                  np.dot(vec3_mn, rho3_mmG))
                                rho3_mmG = rho_0mnG + rho_1mnG
                                vec0_mn = v0_kmn[ikq_k[iK1], mci:mcf, ni:nf]
                                vec1_mn = v1_kmn[ikq_k[iK1], mci:mcf, ni:nf]
                                vec2_mn = v0_kmn[ikq, mci:mcf, ni:nf]
                                vec3_mn = v1_kmn[ikq, mci:mcf, ni:nf]
                                rho_0mnG = np.dot(vec0_mn.conj(),
                                                  np.dot(vec2_mn, rho4_nnG))
                                rho_1mnG = np.dot(vec1_mn.conj(),
                                                  np.dot(vec3_mn, rho4_nnG))
                                rho4_nnG = rho_0mnG + rho_1mnG

                            self.context.timer.start('Screened exchange')
                            W_mnmn = np.einsum(
                                'ijk,km,pqm->ipjq',
                                rho3_mmG.conj(),
                                screened_potential.W_qGG[iq],
                                rho4_nnG,
                                optimize='optimal')
                            W_mnmn *= Ns * so
                            H_ksmnKsmn[ik1, s1, :, :, iK2, s1] -= 0.5 * W_mnmn
                            self.context.timer.stop('Screened exchange')
            if iK1 % (myKsize // 5 + 1) == 0:
                dt = time() - t0
                tleft = dt * myKsize / (iK1 + 1) - dt
                self.context.print(
                    '  Finished %s pair orbitals in %s - Estimated %s left'
                    % ((iK1 + 1) * Nv * Nc * Ns * world.size, timedelta(
                        seconds=round(dt)), timedelta(seconds=round(tleft))))

        H_ksmnKsmn /= self.gs.volume
        self.context.timer.stop('Calculate Hamiltonian')

        mySsize = myKsize * Nv * Nc * Ns
        if myKsize > 0:
            iS0 = myKrange[0] * Nv * Nc * Ns

        df_S = np.reshape(df_Ksmn, -1)
        # multiply by 2 when spin-paired and no SOC
        df_S *= 2.0 / nK / Ns / so
        deps_s = np.reshape(deps_ksmn, -1)
        H_sS = np.reshape(H_ksmnKsmn, (mySsize, self.nS))
        for iS in range(mySsize):
            # Multiply by occupations and adiabatic coupling
            H_sS[iS] *= df_S[iS0 + iS]
            # add bare transition energies
            H_sS[iS, iS0 + iS] += deps_s[iS]

        return BSEMatrix(rhoG0_S, df_S, H_sS)

    @timer('get_density_matrix')
    def get_density_matrix(self, pair_calc, screened_potential, kpt1, kpt2):
        self.context.timer.start('Symop')
        from gpaw.response.g0w0 import QSymmetryOp, get_nmG
        symop, iq = QSymmetryOp.get_symop_from_kpair(self.kd, self.qd,
                                                     kpt1, kpt2)
        qpd = screened_potential.qpd_q[iq]
        nG = qpd.ngmax
        pawcorr0 = screened_potential.pawcorr_q[iq]
        pawcorr, I_G = symop.apply_symop_q(qpd, pawcorr0, kpt1, kpt2)
        self.context.timer.stop('Symop')

        rho_mnG = np.zeros((len(kpt1.eps_n), len(kpt2.eps_n), nG),
                           complex)
        for m in range(len(rho_mnG)):
            rho_mnG[m] = get_nmG(kpt1, kpt2, pawcorr, m, qpd, I_G,
                                 pair_calc, timer=self.context.timer)
        return rho_mnG, iq

    @cached_property
    def _chi0calc(self):
        return Chi0Calculator(
            self.gs, self.context.with_txt('chi0.txt'),
            wd=FrequencyDescriptor([0.0]),
            eta=0.001,
            ecut=self.ecut * Hartree,
            intraband=False,
            hilbert=False,
            nbands=self.nbands)

    @cached_property
    def blockcomm(self):
        return self._chi0calc.chi0_body_calc.blockcomm

    @cached_property
    def wcontext(self):
        return ResponseContext(txt='w.txt', comm=world)

    @cached_property
    def _wcalc(self):
        return initialize_w_calculator(
            self._chi0calc, self.wcontext,
            coulomb=self.coulomb,
            integrate_gamma=self.integrate_gamma)

    @timer('calculate_screened_potential')
    def calculate_screened_potential(self):
        """Calculate W_GG(q)."""

        pawcorr_q = []
        W_qGG = []
        qpd_q = []

        t0 = time()
        self.context.print('Calculating screened potential')
        for iq, q_c in enumerate(self.qd.ibzk_kc):
            chi0 = self._chi0calc.calculate(q_c)
            W_wGG = self._wcalc.calculate_W_wGG(chi0)
            W_GG = W_wGG[0]
            # This is such a terrible way to access the paw
            # corrections. Attributes should not be groped like
            # this... Change in the future! XXX
            pawcorr_q.append(self._chi0calc.chi0_body_calc.pawcorr)
            qpd_q.append(chi0.qpd)
            W_qGG.append(W_GG)

            if iq % (self.qd.nibzkpts // 5 + 1) == 2:
                dt = time() - t0
                tleft = dt * self.qd.nibzkpts / (iq + 1) - dt
                self.context.print(
                    '  Finished {} q-points in {} - Estimated {} left'.format(
                        iq + 1, timedelta(seconds=round(dt)), timedelta(
                            seconds=round(tleft))))

        return ScreenedPotential(pawcorr_q, W_qGG, qpd_q)

    @timer('diagonalize')
    def diagonalize_bse_matrix(self, bsematrix):
        self.context.print('Diagonalizing Hamiltonian')
        if self.use_tammdancoff:
            return bsematrix.diagonalize_tamdancoff(self)
        else:
            return bsematrix.diagonalize_nontamdancoff(self)

    @timer('get_bse_matrix')
    def get_bse_matrix(self, optical=True):
        """Calculate BSE matrix."""
        return self.calculate(optical=optical)

    @timer('get_vchi')
    def get_vchi(self, w_w=None, eta=0.1, optical=True, write_eig=None):
        """Returns v * chi where v is the bare Coulomb interaction"""

        bsematrix = self.get_bse_matrix(optical=optical)

        self.context.print('Calculating response function at %s frequency '
                           'points' % len(w_w))
        vchi_w = np.zeros(len(w_w), dtype=complex)

        w_T, C_T = self.diagonalize_bse_matrix(bsematrix)

        eta /= Hartree
        for iw, w in enumerate(w_w / Hartree):
            tmp_T = 1. / (w - w_T + 1j * eta)
            vchi_w[iw] += np.dot(tmp_T, C_T)
        vchi_w *= 4 * np.pi / self.gs.volume

        if not np.allclose(self.q_c, 0.0):
            cell_cv = self.gs.gd.cell_cv
            B_cv = 2 * np.pi * np.linalg.inv(cell_cv).T
            q_v = np.dot(self.q_c, B_cv)
            vchi_w /= np.dot(q_v, q_v)

        """Check f-sum rule."""
        nv = self.gs.nvalence
        dw_w = (w_w[1:] - w_w[:-1]) / Hartree
        wchi_w = (w_w[1:] * vchi_w[1:] + w_w[:-1] * vchi_w[:-1]) / Hartree / 2
        N = -np.dot(dw_w, wchi_w.imag) * self.gs.volume / (2 * np.pi**2)
        self.context.print('', flush=False)
        self.context.print('Checking f-sum rule:', flush=False)
        self.context.print(f'  Valence = {nv}, N = {N:f}', flush=False)
        self.context.print('')

        if write_eig is not None:
            assert isinstance(write_eig, str)
            filename = write_eig
            if world.rank == 0:
                write_bse_eigenvalues(filename, self.mode,
                                      w_T * Hartree, C_T)

        return vchi_w

    def get_dielectric_function(self, *args, filename='df_bse.csv', **kwargs):
        vchi = self.vchi(*args, optical=True, **kwargs)
        return vchi.dielectric_function(filename=filename)

    def get_eels_spectrum(self, *args, filename='df_bse.csv', **kwargs):
        vchi = self.vchi(*args, optical=False, **kwargs)
        return vchi.eels_spectrum(filename=filename)

    def get_polarizability(self, *args, filename='pol_bse.csv', **kwargs):
        # Previously it was
        # optical = (self.coulomb.truncation is None)
        # I.e. if a truncated kernel is used optical = False.
        # The reason it was set to False with Coulomb
        # truncation is that for q=0 V(G=0) is already
        # set to zero with the truncated coulomb Kernel.
        # However for finite q V(G=0) is different from zero.
        # Therefore the absorption spectra for 2D materials
        # calculated with the previous code was only correct for q=0.
        # See Issue #1055, the related MR and comments therein
        # For simplicity we set it to true for all cases here.
        vchi = self.vchi(*args, optical=True, **kwargs)
        return vchi.polarizability(filename=filename)

    def vchi(self, w_w=None, eta=0.1, write_eig='eig.dat',
             optical=True):
        vchi_w = self.get_vchi(w_w=w_w, eta=eta, optical=optical,
                               write_eig=write_eig)
        return VChi(self.gs.cd, self.context, w_w, vchi_w, optical=optical)

    def collect_A_SS(self, A_sS):
        if world.rank == 0:
            A_SS = np.zeros((self.nS, self.nS), dtype=complex)
            A_SS[:len(A_sS)] = A_sS
            Ntot = len(A_sS)
            for rank in range(1, world.size):
                nkr, nk, ns = self.parallelisation_sizes(rank)
                buf = np.empty((ns, self.nS), dtype=complex)
                world.receive(buf, rank, tag=123)
                A_SS[Ntot:Ntot + ns] = buf
                Ntot += ns
        else:
            world.send(A_sS, 0, tag=123)
        world.barrier()
        if world.rank == 0:
            return A_SS

    def parallelisation_sizes(self, rank=None):
        if rank is None:
            rank = world.rank
        nK = self.kd.nbzkpts
        myKsize = -(-nK // world.size)
        myKrange = range(rank * myKsize,
                         min((rank + 1) * myKsize, nK))
        myKsize = len(myKrange)
        mySsize = myKsize * self.nv * self.nc * self.spins
        mySsize *= (1 + self.spinors)**2
        return myKrange, myKsize, mySsize

    def print_initialization(self, td, eshift, gw_skn):
        isl = ['----------------------------------------------------------',
               f'{self.mode} Hamiltonian',
               '----------------------------------------------------------',
               f'Started at:  {ctime()}', '',
               'Atoms                          : '
               f'{self.gs.atoms.get_chemical_formula(mode="hill")}',
               f'Ground state XC functional     : {self.gs.xcname}',
               f'Valence electrons              : {self.gs.nvalence}',
               f'Spinor calculations            : {self.spinors}',
               f'Number of bands                : {self.gs.bd.nbands}',
               f'Number of spins                : {self.gs.nspins}',
               f'Number of k-points             : {self.kd.nbzkpts}',
               f'Number of irreducible k-points : {self.kd.nibzkpts}',
               f'Number of q-points             : {self.qd.nbzkpts}',
               f'Number of irreducible q-points : {self.qd.nibzkpts}', '']

        for q in self.qd.ibzk_kc:
            isl.append(f'    q: [{q[0]:1.4f} {q[1]:1.4f} {q[2]:1.4f}]')
        isl.append('')
        if gw_skn is not None:
            isl.append('User specified BSE bands')
        isl.extend([f'Response PW cutoff             : {self.ecut * Hartree} '
                    f'eV',
                    f'Screening bands included       : {self.nbands}'])
        if len(self.val_sn) == 1:
            isl.extend([f'Valence bands                  : {self.val_sn[0]}',
                        f'Conduction bands               : {self.con_sn[0]}'])
        else:
            isl.extend([f'Valence bands                  : {self.val_sn[0]}'
                        f' {self.val_sn[1]}',
                        f'Conduction bands               : {self.con_sn[0]}'
                        f' {self.con_sn[1]}'])
        if eshift is not None:
            isl.append(f'Scissors operator              : {eshift * Hartree}'
                       f'eV')
        isl.extend([
            f'Tamm-Dancoff approximation     : {td}',
            f'Number of pair orbitals        : {self.nS}',
            '',
            f'Truncation of Coulomb kernel   : {self.coulomb.truncation}'])
        if self.integrate_gamma == 0:
            isl.append(
                'Coulomb integration scheme     : Analytical - gamma only')
        elif self.integrate_gamma == 1:
            isl.append(
                'Coulomb integration scheme     : Numerical - all q-points')
        else:
            pass
        isl.extend([
            '',
            '----------------------------------------------------------',
            '----------------------------------------------------------',
            '',
            f'Parallelization - Total number of CPUs   : {world.size}',
            '  Screened potential',
            f'    K-point/band decomposition           : {world.size}',
            '  Hamiltonian',
            f'    Pair orbital decomposition           : {world.size}'])
        self.context.print('\n'.join(isl))


class BSE(BSEBackend):
    def __init__(self, calc=None, timer=None, txt='-', **kwargs):
        """Creates the BSE object

        calc: str or calculator object
            The string should refer to the .gpw file contaning KS orbitals
        ecut: float
            Plane wave cutoff energy (eV)
        nbands: int
            Number of bands used for the screened interaction
        valence_bands: list
            Valence bands used in the BSE Hamiltonian
        conduction_bands: list
            Conduction bands used in the BSE Hamiltonian
        eshift: float
            Scissors operator opening the gap (eV)
         q_c: list of three floats
-            Wavevector in reduced units on which the response is calculated
        direction: int
            if q_c = [0, 0, 0] this gives the direction in cartesian
            coordinates - 0=x, 1=y, 2=z
        gw_skn: list / array
            List or array defining the gw quasiparticle energies used in
            the BSE Hamiltonian. Should match spin, k-points and
            valence/conduction bands
        truncation: str or None
            Coulomb truncation scheme. Can be None or 2D.
        integrate_gamma: int
            Method to integrate the Coulomb interaction. 1 is a numerical
            integration at all q-points with G=[0,0,0] - this breaks the
            symmetry slightly. 0 is analytical integration at q=[0,0,0] only -
            this conserves the symmetry. integrate_gamma=2 is the same as 1,
            but the average is only carried out in the non-periodic directions.
        txt: str
            txt output
        mode: str
            Theory level used. can be RPA TDHF or BSE. Only BSE is screened.
        """
        gs, context = get_gs_and_context(
            calc, txt, world=world, timer=timer)

        super().__init__(gs=gs, context=context, **kwargs)


def write_bse_eigenvalues(filename, mode, w_w, C_w):
    with open(filename, 'w') as fd:
        print('# %s eigenvalues (in eV) and weights' % mode, file=fd)
        print('# Number   eig   weight', file=fd)
        for iw, (w, C) in enumerate(zip(w_w, C_w)):
            print('%8d %12.6f %12.16f' % (iw, w.real, C.real),
                  file=fd)


def read_bse_eigenvalues(filename):
    _, w_w, C_w = np.loadtxt(filename, unpack=True)
    return w_w, C_w


def write_spectrum(filename, w_w, A_w):
    with open(filename, 'w') as fd:
        for w, A in zip(w_w, A_w):
            print(f'{w:.9f}, {A:.9f}', file=fd)


def read_spectrum(filename):
    w_w, A_w = np.loadtxt(filename, delimiter=',',
                          unpack=True)
    return w_w, A_w


@dataclass
class VChi:
    cd: CellDescriptor
    context: ResponseContext
    w_w: np.ndarray
    vchi_w: np.ndarray
    optical: bool

    def epsilon(self):
        assert self.optical
        return -self.vchi_w + 1.0

    def eels(self):
        assert not self.optical
        return -self.vchi_w.imag

    def alpha(self):
        assert self.optical
        L = self.cd.nonperiodic_hypervolume
        return -L * self.vchi_w / (4 * np.pi)

    def dielectric_function(self, filename='df_bse.csv'):
        """Returns and writes real and imaginary part of the dielectric
        function.

        w_w: list of frequencies (eV)
            Dielectric function is calculated at these frequencies
        eta: float
            Lorentzian broadening of the spectrum (eV)
        filename: str
            data file on which frequencies, real and imaginary part of
            dielectric function is written
        write_eig: str
            File on which the BSE eigenvalues are written
        """

        return self._hackywrite(self.epsilon(), filename)

    # XXX The default filename clashes with that of dielectric function!
    def eels_spectrum(self, filename='df_bse.csv'):
        """Returns and writes real and imaginary part of the dielectric
        function.

        w_w: list of frequencies (eV)
            Dielectric function is calculated at these frequencies
        eta: float
            Lorentzian broadening of the spectrum (eV)
        filename: str
            data file on which frequencies, real and imaginary part of
            dielectric function is written
        write_eig: str
            File on which the BSE eigenvalues are written
        """
        return self._hackywrite(self.eels(), filename)

    def polarizability(self, filename='pol_bse.csv'):
        r"""Calculate the polarizability alpha.
        In 3D the imaginary part of the polarizability is related to the
        dielectric function by Im(eps_M) = 4 pi * Im(alpha). In systems
        with reduced dimensionality the converged value of alpha is
        independent of the cell volume. This is not the case for eps_M,
        which is ill defined. A truncated Coulomb kernel will always give
        eps_M = 1.0, whereas the polarizability maintains its structure.
        pbs should be a list of booleans giving the periodic directions.

        By default, generate a file 'pol_bse.csv'. The three colomns are:
        frequency (eV), Real(alpha), Imag(alpha). The dimension of alpha
        is \AA to the power of non-periodic directions.
        """
        return self._hackywrite(self.alpha(), filename)

    def _hackywrite(self, array, filename):
        if world.rank == 0 and filename is not None:
            if array.dtype == complex:
                write_response_function(filename, self.w_w, array.real,
                                        array.imag)
            else:
                assert array.dtype == float
                write_spectrum(filename, self.w_w, array)

        world.barrier()

        self.context.print('Calculation completed at:', ctime(), flush=False)
        self.context.print('')

        return self.w_w, array



from ase import Atom, Atoms
from gpaw import GPAW
import pytest


def test_eigen_cg():
    a = 4.05
    d = a / 2**0.5
    bulk = Atoms([Atom('Al', (0, 0, 0)),
                  Atom('Al', (0.5, 0.5, 0.5))],
                 pbc=True)
    bulk.set_cell((d, d, a), scale_atoms=True)
    h = 0.25
    base_params = dict(
        mode='fd',
        h=h,
        nbands=2 * 8,
        kpts=(2, 2, 2),
        convergence={'energy': 1e-5})
    calc = GPAW(**base_params)
    bulk.calc = calc
    e0 = bulk.get_potential_energy()
    calc = GPAW(**base_params, eigensolver='cg')
    bulk.calc = calc
    e1 = bulk.get_potential_energy()
    assert e0 == pytest.approx(e1, abs=5.e-5)

    energy_tolerance = 0.001
    assert e0 == pytest.approx(-6.97626, abs=energy_tolerance)
    assert e1 == pytest.approx(-6.97627, abs=energy_tolerance)



import pytest
from gpaw import GPAW
from gpaw.xc.rpa import RPACorrelation
from gpaw.mpi import serial_comm


@pytest.mark.rpa
@pytest.mark.response
def test_rpa_rpa_energy_Na(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['na_pw'], communicator=serial_comm)
    ecut = 120
    rpa = RPACorrelation(calc, txt=f'rpa_{ecut}s.txt', ecut=[ecut])
    E = rpa.calculate()
    assert E == pytest.approx(-1.106, abs=0.005)



from pathlib import Path

from gpaw.new.ase_interface import GPAW
from gpaw.nlopt.matrixel import make_nlodata
from gpaw.test.cachedfilehandler import CachedFilesHandler
from gpaw.test.gpwfile import GPWFiles, response_band_cutoff


class MMEFiles(CachedFilesHandler):
    """Create files that store momentum matrix elements."""

    def __init__(self, folder: Path, gpw_files: GPWFiles):
        super().__init__(folder, '.npz')
        self.gpw_files = gpw_files

    def _calculate_and_write(self, name, work_path):
        calc = GPAW(self.gpw_files[name], parallel={'domain': 1, 'band': 1})
        nb = response_band_cutoff[
            name if not name.endswith('_spinpol') else name[:-8]]
        nlodata = make_nlodata(calc, ni=0, nf=nb)
        nlodata.write(work_path)



from gpaw.new.ibzwfs import IBZWaveFunctions


class PWFDIBZWaveFunction(IBZWaveFunctions):
    def move(self, fracpos_ac, atomdist):
        super().move(fracpos_ac, atomdist)
        for wfs in self:
            wfs.move(fracpos_ac, atomdist)



import pytest
import numpy as np
from ase.build import molecule
from gpaw import GPAW
from gpaw.mpi import world, serial_comm, broadcast_exception

from gpaw.test.lcaotddft.test_molecule import only_on_master

pytestmark = [pytest.mark.usefixtures('module_tmp_path'),
              pytest.mark.gllb,
              pytest.mark.libxc
              ]


def check_asp(ref_asp, D_asp, atol):
    assert ref_asp is not D_asp, \
        'Trying to compare same objects. Is the test broken?'
    assert np.allclose(ref_asp.toarray(), D_asp.toarray(), atol=atol, rtol=0)


def check_response(ref_response, response, atol=1e-12):
    # Collect response in master
    D_asp = response.D_asp.copy()
    D_asp.redistribute(D_asp.partition.as_serial())
    Dresp_asp = response.Dresp_asp.copy()
    Dresp_asp.redistribute(Dresp_asp.partition.as_serial())
    vt_sG = response.gd.collect(response.vt_sG)

    with broadcast_exception(world):
        if world.rank == 0:
            check_asp(ref_response.D_asp, D_asp, atol=atol)
            check_asp(ref_response.Dresp_asp, Dresp_asp, atol=atol)
            assert np.allclose(ref_response.vt_sG, vt_sG, rtol=0, atol=atol)


def check_partitions(response, gs_calc):
    partition = response.D_asp.partition
    assert response.Dresp_asp.partition is partition
    assert partition == gs_calc.wfs.atom_partition
    assert partition == gs_calc.density.atom_partition


def calculate_ground_state(**kwargs):
    atoms = molecule('H2O')
    atoms.center(vacuum=4)
    gs_calc = GPAW(mode='lcao', basis='sz(dzp)', h=0.4,
                   xc='GLLBSC',
                   txt='gs.out',
                   **kwargs)
    atoms.calc = gs_calc
    atoms.get_potential_energy()
    return gs_calc


@pytest.fixture(scope='module')
@only_on_master(world)
def ground_state_calculation():
    gs_calc = calculate_ground_state(communicator=serial_comm)
    gs_calc.write('gs.gpw', mode='all')
    return gs_calc.hamiltonian.xc.response


def test_read(ground_state_calculation, module_tmp_path):
    ref_response = ground_state_calculation

    gs_calc = GPAW(module_tmp_path / 'gs.gpw', txt=None)
    response = gs_calc.hamiltonian.xc.response
    check_response(ref_response, response)
    check_partitions(response, gs_calc)


def test_calculate(ground_state_calculation, in_tmp_dir):
    ref_response = ground_state_calculation

    parallel = {'band': 2 if world.size >= 4 else 1}
    gs_calc = calculate_ground_state(parallel=parallel)
    response = gs_calc.hamiltonian.xc.response
    check_response(ref_response, response)
    check_partitions(response, gs_calc)


def test_fixed_density(ground_state_calculation, module_tmp_path, in_tmp_dir):
    ref_response = ground_state_calculation

    parallel = {'band': 2 if world.size >= 4 else 1}
    bs_calc = GPAW(module_tmp_path / 'gs.gpw', txt=None) \
        .fixed_density(parallel=parallel,
                       txt='unocc.out',
                       )
    response = bs_calc.hamiltonian.xc.response
    check_response(ref_response, response)
    # check_partitions(response, bs_calc)  # Fails. Should it work?
    bs_calc.write('unocc.gpw', mode='all')

    bs_calc = GPAW('unocc.gpw', txt=None)
    response = bs_calc.hamiltonian.xc.response
    check_response(ref_response, response)
    check_partitions(response, bs_calc)



"""
Potentials for orbital density dependent energy functionals
"""
from ase.units import Hartree
import numpy as np

from gpaw.directmin.tools import d_matrix
from gpaw.lfc import LFC
from gpaw.poisson import PoissonSolver
from gpaw.transformers import Transformer
from gpaw.utilities import pack_density, unpack_hermitian


class PZSICLCAO:
    """
    Perdew-Zunger self-interaction corrections

    """
    def __init__(self, wfs, dens, ham, scaling_factor=(1.0, 1.0),
                 sic_coarse_grid=True, store_potentials=False,
                 poisson_solver='FPS'):

        self.name = 'PZ-SIC'
        # what we need from wfs
        self.setups = wfs.setups
        self.bfs = wfs.basis_functions
        spos_ac = wfs.spos_ac
        self.cgd = wfs.gd

        # what we need from dens
        self.finegd = dens.finegd
        self.sic_coarse_grid = sic_coarse_grid

        if self.sic_coarse_grid:
            self.ghat_cg = LFC(self.cgd,
                               [setup.ghat_l for setup
                                in self.setups],
                               integral=np.sqrt(4 * np.pi),
                               forces=True)
            self.ghat_cg.set_positions(spos_ac)
            self.ghat = None
        else:
            self.ghat = dens.ghat  # we usually solve poiss. on finegd
            self.ghat_cg = None

        # what we need from ham
        self.xc = ham.xc

        if poisson_solver == 'FPS':
            self.poiss = PoissonSolver(use_charge_center=True,
                                       use_charged_periodic_corrections=True)
        elif poisson_solver == 'GS':
            self.poiss = PoissonSolver(name='fd',
                                       relax=poisson_solver,
                                       eps=1.0e-16,
                                       use_charge_center=True,
                                       use_charged_periodic_corrections=True)

        if self.sic_coarse_grid is True:
            self.poiss.set_grid_descriptor(self.cgd)
        else:
            self.poiss.set_grid_descriptor(self.finegd)

        self.interpolator = Transformer(self.cgd, self.finegd, 3)
        self.restrictor = Transformer(self.finegd, self.cgd, 3)
        self.dtype = wfs.dtype
        self.eigv_s = {}
        self.lagr_diag_s = {}
        self.e_sic_by_orbitals = {}
        self.counter = 0  # number of calls of this class
        # Scaling factor:
        self.beta_c = scaling_factor[0]
        self.beta_x = scaling_factor[1]

        self.n_kps = wfs.kd.nibzkpts
        self.store_potentials = store_potentials
        if store_potentials:
            self.old_pot = {}
            for kpt in wfs.kpt_u:
                k = self.n_kps * kpt.s + kpt.q
                n_occ = 0
                nbands = len(kpt.f_n)
                while n_occ < nbands and kpt.f_n[n_occ] > 1e-10:
                    n_occ += 1
                self.old_pot[k] = self.cgd.zeros(n_occ, dtype=float)

    def get_gradients(self, h_mm, c_nm, f_n, evec, evals, kpt,
                      wfs, timer, matrix_exp, repr_name,
                      ind_up, constraints, occupied_only=False):
        """
        :param C_nM: coefficients of orbitals
        :return: matrix G - gradients, and orbital SI energies

        which is G_{ij} = (1-delta_{ij}/2)*(int_0^1 e^{tA} L e^{-tA} dt )_{ji}

        Lambda_ij = (C_i, F_j C_j )

        L_{ij} = Lambda_ji^{cc} - Lambda_ij

        """

        # 0.
        n_occ = 0
        nbands = len(f_n)
        while n_occ < nbands and f_n[n_occ] > 1e-10:
            n_occ += 1

        if occupied_only is True:
            nbs = n_occ
        else:
            nbs = c_nm.shape[0]
        n_set = c_nm.shape[1]

        timer.start('Construct Gradient Matrix')
        hc_mn = np.dot(h_mm.conj(), c_nm[:nbs].T)
        h_mm = np.dot(c_nm[:nbs].conj(), hc_mn)
        # odd part
        b_mn = np.zeros(shape=(n_set, nbs), dtype=self.dtype)
        e_total_sic = np.array([])
        for n in range(n_occ):
            F_MM, sic_energy_n =\
                self.get_orbital_potential_matrix(f_n, c_nm, kpt,
                                                  wfs, wfs.setups,
                                                  n, timer)

            b_mn[:, n] = np.dot(c_nm[n], F_MM.conj()).T
            e_total_sic = np.append(e_total_sic, sic_energy_n, axis=0)
        l_odd = np.dot(c_nm[:nbs].conj(), b_mn)

        f = f_n[:nbs]
        grad = f * (h_mm[:nbs, :nbs] + l_odd) - \
            f[:, np.newaxis] * (h_mm[:nbs, :nbs] + l_odd.T.conj())

        if matrix_exp in ['pade-approx', 'egdecomp2']:
            grad = np.ascontiguousarray(grad)
        elif matrix_exp == 'egdecomp':
            timer.start('Use Eigendecomposition')
            grad = np.dot(evec.T.conj(), np.dot(grad, evec))
            grad = grad * d_matrix(evals)
            grad = np.dot(evec, np.dot(grad, evec.T.conj()))
            for i in range(grad.shape[0]):
                grad[i][i] *= 0.5
            timer.stop('Use Eigendecomposition')
        else:
            raise NotImplementedError('Check the keyword '
                                      'for matrix_exp. \n'
                                      'Must be '
                                      '\'pade-approx\' or '
                                      '\'egdecomp\'')
        if self.dtype == float:
            grad = grad.real
        if repr_name in ['sparse', 'u-invar']:
            grad = grad[ind_up]

        timer.stop('Construct Gradient Matrix')

        u = kpt.s * self.n_kps + kpt.q
        self.e_sic_by_orbitals[u] = \
            e_total_sic.reshape(e_total_sic.shape[0] // 2, 2)

        timer.start('Residual')
        hc_mn += b_mn
        h_mm += l_odd

        rhs2 = kpt.S_MM.conj() @ c_nm[:n_occ].T @ h_mm[:n_occ, :n_occ]
        hc_mn = hc_mn[:, :n_occ] - rhs2[:, :n_occ]

        if constraints:
            # Zero out the components of the residual that are constrained,
            # so that the constrained degrees of freedom are always considered
            # converged
            for con in constraints:
                con1 = con[0]
                hc_mn[:, con1] = 0.0

        norm = []
        for i in range(n_occ):
            norm.append(np.dot(hc_mn[:, i].conj(),
                               hc_mn[:, i]).real * kpt.f_n[i])

        error = sum(norm) * Hartree ** 2 / wfs.nvalence
        del rhs2, hc_mn, norm
        timer.stop('Residual')

        if self.counter == 0:
            h_mm = 0.5 * (h_mm + h_mm.T.conj())
            kpt.eps_n[:nbs] = np.linalg.eigh(h_mm)[0]
        self.counter += 1

        if constraints:
            constrain_grad(grad, constraints, ind_up)

        return 2.0 * grad, error

    def get_orbital_potential_matrix(self, f_n, C_nM, kpt,
                                     wfs, setup, m, timer):
        """
        :param f_n:
        :param C_nM:
        :param kpt:
        :param wfs:
        :param setup:
        :return: F_i = <m|v_i + PAW_i|n > for occupied
                 F_i = 0 for unoccupied,
                 SI energies

        To calculate this, we need to calculate
        orbital-depended potential and PAW corrections to it.

        Fot this, we need firstly to get orbitals densities.

        """
        kpoint = self.n_kps * kpt.s + kpt.q
        n_set = C_nM.shape[1]
        F_MM = np.zeros(shape=(n_set, n_set),
                        dtype=self.dtype)
        # get orbital-density
        timer.start('Construct Density, Charge, adn DM')
        nt_G, Q_aL, D_ap = \
            self.get_density(f_n,
                             C_nM, kpt,
                             wfs, setup, m)
        timer.stop('Construct Density, Charge, adn DM')

        # calculate sic energy,
        # sic pseudo-potential and Hartree
        timer.start('Get Pseudo Potential')
        e_sic_m, vt_mG, vHt_g = \
            self.get_pseudo_pot(nt_G, Q_aL, m, kpoint, timer)
        timer.stop('Get Pseudo Potential')

        # calculate PAW corrections
        timer.start('PAW')
        e_sic_paw_m, dH_ap = \
            self.get_paw_corrections(D_ap, vHt_g, timer)
        timer.stop('PAW')

        # total sic:
        e_sic_m += e_sic_paw_m

        timer.start('ODD Potential Matrices')
        Vt_MM = np.zeros_like(F_MM)
        self.bfs.calculate_potential_matrix(vt_mG, Vt_MM, kpt.q)
        # make matrix hermitian
        ind_l = np.tril_indices(Vt_MM.shape[0], -1)
        Vt_MM[(ind_l[1], ind_l[0])] = Vt_MM[ind_l].conj()
        timer.stop('ODD Potential Matrices')

        # PAW:
        timer.start('Potential matrix - PAW')
        for a, dH_p in dH_ap.items():
            P_Mj = wfs.P_aqMi[a][kpt.q]
            dH_ij = unpack_hermitian(dH_p)

            if self.dtype is complex:
                F_MM += P_Mj @ dH_ij @ P_Mj.T.conj()
            else:
                F_MM += P_Mj @ dH_ij @ P_Mj.T

        if self.dtype is complex:
            F_MM += Vt_MM.astype(complex)
        else:
            F_MM += Vt_MM
        if self.sic_coarse_grid:
            self.cgd.comm.sum(F_MM)
        else:
            self.finegd.comm.sum(F_MM)
        timer.stop('Potential matrix - PAW')

        return F_MM, e_sic_m * f_n[m]

    def get_density(self, f_n, C_nM, kpt,
                    wfs, setup, m):

        # construct orbital density matrix
        if f_n[m] > 1.0 + 1.0e-4:
            occup_factor = f_n[m] / (3.0 - wfs.nspins)
        else:
            occup_factor = f_n[m]
        rho_MM = occup_factor * np.outer(C_nM[m].conj(), C_nM[m])

        nt_G = self.cgd.zeros()
        self.bfs.construct_density(rho_MM, nt_G, kpt.q)

        # calculate  atomic density matrix and
        # compensation charges
        D_ap = {}
        Q_aL = {}

        for a in wfs.P_aqMi.keys():
            P_Mi = wfs.P_aqMi[a][kpt.q]
            D_ii = np.zeros((wfs.P_aqMi[a].shape[2],
                             wfs.P_aqMi[a].shape[2]),
                            dtype=self.dtype)
            rhoP_Mi = rho_MM @ P_Mi
            D_ii = P_Mi.T.conj() @ rhoP_Mi
            if self.dtype is complex:
                D_ap[a] = D_p = pack_density(D_ii.real)
            else:
                D_ap[a] = D_p = pack_density(D_ii)

            Q_aL[a] = np.dot(D_p, setup[a].Delta_pL)

        return nt_G, Q_aL, D_ap

    def get_pseudo_pot(self, nt, Q_aL, i, kpoint, timer):

        if self.sic_coarse_grid is False:
            # change to fine grid
            vt_sg = self.finegd.zeros(2)
            vHt_g = self.finegd.zeros()
            nt_sg = self.finegd.zeros(2)
        else:
            vt_sg = self.cgd.zeros(2)
            vHt_g = self.cgd.zeros()
            nt_sg = self.cgd.zeros(2)

        if self.sic_coarse_grid is False:
            self.interpolator.apply(nt, nt_sg[0])
            nt_sg[0] *= self.cgd.integrate(nt) / \
                self.finegd.integrate(nt_sg[0])
        else:
            nt_sg[0] = nt

        timer.start('ODD XC 3D grid')
        if self.sic_coarse_grid is False:
            e_xc = self.xc.calculate(self.finegd, nt_sg, vt_sg)
        else:
            e_xc = self.xc.calculate(self.cgd, nt_sg, vt_sg)
        timer.stop('ODD XC 3D grid')
        vt_sg[0] *= -self.beta_x

        # Hartree
        if self.sic_coarse_grid is False:
            self.ghat.add(nt_sg[0], Q_aL)
        else:
            self.ghat_cg.add(nt_sg[0], Q_aL)

        timer.start('ODD Poisson')
        if self.store_potentials:
            if self.sic_coarse_grid:
                vHt_g = self.old_pot[kpoint][i]
            else:
                self.interpolator.apply(self.old_pot[kpoint][i],
                                        vHt_g)
        self.poiss.solve(vHt_g, nt_sg[0],
                         zero_initial_phi=self.store_potentials,
                         timer=timer)
        if self.store_potentials:
            if self.sic_coarse_grid:
                self.old_pot[kpoint][i] = vHt_g.copy()
            else:
                self.restrictor.apply(vHt_g, self.old_pot[kpoint][i])

        timer.stop('ODD Poisson')

        timer.start('ODD Hartree integrate')
        if self.sic_coarse_grid is False:
            ec = 0.5 * self.finegd.integrate(nt_sg[0] * vHt_g)
        else:
            ec = 0.5 * self.cgd.integrate(nt_sg[0] * vHt_g)

        timer.stop('ODD Hartree integrate')
        vt_sg[0] -= vHt_g * self.beta_c
        if self.sic_coarse_grid is False:
            vt_G = self.cgd.zeros()
            self.restrictor.apply(vt_sg[0], vt_G)
        else:
            vt_G = vt_sg[0]

        return np.array([-ec * self.beta_c,
                         -e_xc * self.beta_x]), vt_G, vHt_g

    def get_paw_corrections(self, D_ap, vHt_g, timer):

        # XC-PAW
        timer.start('xc-PAW')
        dH_ap = {}
        exc = 0.0
        for a, D_p in D_ap.items():
            setup = self.setups[a]
            dH_sp = np.zeros((2, len(D_p)))
            D_sp = np.array([D_p, np.zeros_like(D_p)])
            exc += self.xc.calculate_paw_correction(setup, D_sp,
                                                    dH_sp,
                                                    addcoredensity=False,
                                                    a=a)
            dH_ap[a] = -dH_sp[0] * self.beta_x
        timer.stop('xc-PAW')

        # Hartree-PAW
        timer.start('Hartree-PAW')
        ec = 0.0
        timer.start('ghat-PAW')
        if self.sic_coarse_grid is False:
            W_aL = self.ghat.dict()
            self.ghat.integrate(vHt_g, W_aL)
        else:
            W_aL = self.ghat_cg.dict()
            self.ghat_cg.integrate(vHt_g, W_aL)
        timer.stop('ghat-PAW')

        for a, D_p in D_ap.items():
            setup = self.setups[a]
            M_p = np.dot(setup.M_pp, D_p)
            ec += np.dot(D_p, M_p)
            dH_ap[a] += -(2.0 * M_p + np.dot(setup.Delta_pL,
                                             W_aL[a])) * self.beta_c
        timer.stop('Hartree-PAW')

        timer.start('Wait for sum')
        if self.sic_coarse_grid is False:
            ec = self.finegd.comm.sum_scalar(ec)
            exc = self.finegd.comm.sum_scalar(exc)
        else:
            ec = self.cgd.comm.sum_scalar(ec)
            exc = self.cgd.comm.sum_scalar(exc)
        timer.stop('Wait for sum')

        return np.array([-ec * self.beta_c, -exc * self.beta_x]), dH_ap

    def get_odd_corrections_to_forces(self, wfs, dens):

        timer = wfs.timer

        timer.start('ODD corrections')

        natoms = len(wfs.setups)
        F_av = np.zeros((natoms, 3))
        Ftheta_av = np.zeros_like(F_av)
        Frho_av = np.zeros_like(F_av)
        Fatom_av = np.zeros_like(F_av)
        Fpot_av = np.zeros_like(F_av)
        Fhart_av = np.zeros_like(F_av)

        ksl = wfs.ksl
        nao = ksl.nao
        mynao = ksl.mynao
        dtype = wfs.dtype
        manytci = wfs.manytci
        gd = wfs.gd

        Mstart = ksl.Mstart
        Mstop = ksl.Mstop
        n_kps = wfs.kd.nibzkpts

        timer.start('TCI derivative')
        dThetadR_qvMM, dTdR_qvMM = manytci.O_qMM_T_qMM(
            gd.comm, Mstart, Mstop, False, derivative=True)

        dPdR_aqvMi = manytci.P_aqMi(
            self.bfs.my_atom_indices, derivative=True)
        gd.comm.sum(dThetadR_qvMM)
        gd.comm.sum(dTdR_qvMM)
        timer.stop('TCI derivative')

        my_atom_indices = self.bfs.my_atom_indices
        atom_indices = self.bfs.atom_indices

        def _slices(indices):
            for a in indices:
                M1 = self.bfs.M_a[a] - Mstart
                M2 = M1 + self.setups[a].nao
                if M2 > 0:
                    yield a, max(0, M1), M2

        def slices():
            return _slices(atom_indices)

        def my_slices():
            return _slices(my_atom_indices)

        #
        #         -----                    -----
        #          \    -1                  \    *
        # E      =  )  S     H    rho     =  )  c     eps  f  c
        #  mu nu   /    mu x  x z    z nu   /    n mu    n  n  n nu
        #         -----                    -----
        #          x z                       n
        #
        # We use the transpose of that matrix.  The first form is used
        # if rho is given, otherwise the coefficients are used.

        for kpt in wfs.kpt_u:
            u = kpt.s * n_kps + kpt.q
            f_n = kpt.f_n
            n_occ = 0
            for f in f_n:
                if f > 1.0e-10:
                    n_occ += 1

            for m in range(n_occ):

                # calculate orbital-density matrix
                rho_xMM = kpt.f_n[m] * np.outer(
                    kpt.C_nM[m].conj(), kpt.C_nM[m]) / (3.0 - wfs.nspins)

                F_MM = self.get_orbital_potential_matrix(
                    f_n, kpt.C_nM, kpt, wfs, self.setups, m, timer)[0]

                sfrhoT_MM = np.linalg.solve(
                    wfs.S_qMM[kpt.q], F_MM @ rho_xMM).T.copy()

                del F_MM

                # Density matrix contribution due to basis overlap
                #
                #            ----- d Theta
                #  a          \           mu nu
                # F  += -2 Re  )   ------------  E
                #             /        d R        nu mu
                #            -----        mu nu
                #         mu in a; nu
                #

                dThetadRE_vMM = (dThetadR_qvMM[kpt.q] *
                                 sfrhoT_MM[np.newaxis]).real
                for a, M1, M2 in my_slices():
                    Ftheta_av[a, :] += \
                        -2.0 * dThetadRE_vMM[:, M1:M2].sum(-1).sum(-1)
                del dThetadRE_vMM

                # Density matrix contribution from PAW correction
                #
                #           -----                        -----
                #  a         \      a                     \     b
                # F +=  2 Re  )    Z      E        - 2 Re  )   Z      E
                #            /      mu nu  nu mu          /     mu nu  nu mu
                #           -----                        -----
                #           mu nu                    b; mu in a; nu
                #
                # with
                #                  b*
                #         -----  dP
                #   b      \       i mu    b   b
                #  Z     =  )   -------- dS   P
                #   mu nu  /     dR        ij  j nu
                #         -----    b mu
                #           ij
                #
                work_MM = np.zeros((mynao, nao), dtype)
                ZE_MM = None
                for b in my_atom_indices:
                    setup = self.setups[b]
                    dO_ii = np.asarray(setup.dO_ii, dtype)
                    dOP_iM = dO_ii @ wfs.P_aqMi[b][kpt.q].T.conj()

                    for v in range(3):
                        work_MM = \
                            dPdR_aqvMi[b][kpt.q][v][Mstart:Mstop] @ dOP_iM
                        ZE_MM = (work_MM * sfrhoT_MM).real
                        for a, M1, M2 in slices():
                            dE = 2 * ZE_MM[M1:M2].sum()
                            Frho_av[a, v] -= dE  # the "b; mu in a; nu" term
                            Frho_av[b, v] += dE  # the "mu nu" term
                del work_MM, ZE_MM

                # Potential contribution
                #
                #           -----      /  d Phi  (r)
                #  a         \        |        mu    ~
                # F += -2 Re  )       |   ---------- v (r)  Phi  (r) dr rho
                #            /        |     d R                nu         nu mu
                #           -----    /         a
                #        mu in a; nu
                #

                nt_G, Q_aL, D_ap = \
                    self.get_density(f_n, kpt.C_nM, kpt, wfs, self.setups, m)

                e_sic_m, vt_mG, vHt_g = \
                    self.get_pseudo_pot(nt_G, Q_aL, m, u, timer)
                e_sic_paw_m, dH_ap = \
                    self.get_paw_corrections(D_ap, vHt_g, timer)

                Fpot_av += \
                    self.bfs.calculate_force_contribution(
                        vt_mG, rho_xMM.T, kpt.q)

                # Atomic density contribution
                #            -----                         -----
                #  a          \     a                       \   b
                # F  += -2 Re  )   A      rho       + 2 Re   ) A rho
                #             /     mu nu    nu mu          / mununumu
                #            -----                         -----
                #            mu nu                     b; mu in a; nu
                #
                #                  b*
                #         ----- d P
                #  b       \       i mu   b   b
                # A     =   )   ------- dH   P
                #  mu nu   /    d R       ij  j nu
                #         -----    b mu
                #           ij
                #
                for b in my_atom_indices:
                    H_ii = np.asarray(unpack_hermitian(dH_ap[b]),
                                      dtype)
                    HP_iM = H_ii @ wfs.P_aqMi[b][kpt.q].T.conj()
                    for v in range(3):
                        dPdR_Mi = \
                            dPdR_aqvMi[b][kpt.q][v][Mstart:Mstop]
                        ArhoT_MM = \
                            ((dPdR_Mi @ HP_iM) * rho_xMM.T).real
                        for a, M1, M2 in slices():
                            dE = 2 * ArhoT_MM[M1:M2].sum()
                            Fatom_av[a, v] += dE  # the "b; mu in a; nu" term
                            Fatom_av[b, v] -= dE  # the "mu nu" term

                # contribution from hartree
                if self.sic_coarse_grid is False:
                    ghat_aLv = dens.ghat.dict(derivative=True)

                    dens.ghat.derivative(vHt_g, ghat_aLv)
                    for a, dF_Lv in ghat_aLv.items():
                        Fhart_av[a] -= self.beta_c * np.dot(Q_aL[a], dF_Lv)
                else:
                    ghat_aLv = self.ghat_cg.dict(derivative=True)

                    self.ghat_cg.derivative(vHt_g, ghat_aLv)
                    for a, dF_Lv in ghat_aLv.items():
                        Fhart_av[a] -= self.beta_c * np.dot(Q_aL[a], dF_Lv)

        F_av += Fpot_av + Ftheta_av + Frho_av + Fatom_av + Fhart_av

        wfs.gd.comm.sum(F_av, 0)

        timer.start('Wait for sum')
        ksl.orbital_comm.sum(F_av)
        if wfs.bd.comm.rank == 0:
            wfs.kd.comm.sum(F_av, 0)

        wfs.world.broadcast(F_av, 0)
        timer.stop('Wait for sum')

        timer.stop('ODD corrections')

        return F_av * (3.0 - wfs.nspins)

    def get_lagrange_matrices(self, h_mm, c_nm, f_n, kpt,
                              wfs, occupied_only=False,
                              update_eigenvalues=False):
        n_occ = 0
        nbands = len(f_n)
        while n_occ < nbands and f_n[n_occ] > 1e-10:
            n_occ += 1

        if occupied_only is True:
            nbs = n_occ
        else:
            nbs = c_nm.shape[0]
        n_set = c_nm.shape[1]

        hc_mn = np.dot(h_mm.conj(), c_nm[:nbs].T)
        h_mm = np.dot(c_nm[:nbs].conj(), hc_mn)
        # odd part
        b_mn = np.zeros(shape=(n_set, nbs), dtype=self.dtype)
        e_total_sic = np.array([])
        for n in range(n_occ):
            F_MM, sic_energy_n =\
                self.get_orbital_potential_matrix(f_n, c_nm, kpt,
                                                  wfs, wfs.setups,
                                                  n, wfs.timer)

            b_mn[:, n] = np.dot(c_nm[n], F_MM.conj()).T
            e_total_sic = np.append(e_total_sic, sic_energy_n, axis=0)
        l_odd = np.dot(c_nm[:nbs].conj(), b_mn)

        k = wfs.eigensolver.kpointval(kpt)

        fullham = h_mm + 0.5 * (l_odd + l_odd.T.conj())
        fullham[:n_occ, n_occ:] = 0.0
        fullham[n_occ:, :n_occ] = 0.0

        self.lagr_diag_s[k] = np.diagonal(fullham).real

        if update_eigenvalues:
            eigval, eigvec = np.linalg.eigh(fullham[0:n_occ, 0:n_occ])
            kpt.eps_n[0:n_occ] = eigval
            eigval, eigvec = np.linalg.eigh(fullham[n_occ:nbs, n_occ:nbs])
            kpt.eps_n[n_occ:nbs] = eigval

        return h_mm, l_odd


def constrain_grad(grad, constraints, ind_up):
    """
    Zero out the components of the gradient that are constrained, so that no
    optimization step is taken along the constrained degrees of freedom (It
    would be better not to evaluate these components of the gradient to begin
    with.).
    """

    for con in constraints:
        num = -1
        for ind1, ind2 in zip(ind_up[0], ind_up[1]):
            num += 1
            if con == [ind1, ind2]:
                grad[num] = 0.0
    return grad



import pytest

from gpaw import GPAW, LCAO
from gpaw.directmin.etdm_lcao import LCAOETDM
from gpaw.directmin.tools import excite
from gpaw.directmin.derivatives import Davidson
from gpaw.mom import prepare_mom_calculation
from ase import Atoms
import numpy as np


@pytest.mark.later
@pytest.mark.sic
def test_gmf_lcaosic(in_tmp_dir):
    """
    test Perdew-Zunger Self-Interaction
    Correction  in LCAO mode using DirectMin
    :param in_tmp_dir:
    :return:
    """

    # Water molecule:
    d = 0.9575
    t = np.pi / 180 * 104.51
    H2O = Atoms('OH2',
                positions=[(0, 0, 0),
                           (d, 0, 0),
                           (d * np.cos(t), d * np.sin(t), 0)])
    H2O.center(vacuum=3.0)

    calc = GPAW(mode=LCAO(),
                basis='sz(dzp)',
                h=0.24,
                occupations={'name': 'fixed-uniform'},
                eigensolver='etdm-lcao',
                convergence={'eigenstates': 1e-4},
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                spinpol=True,
                symmetry='off'
                )
    H2O.calc = calc
    H2O.get_potential_energy()

    calc.set(eigensolver=LCAOETDM(excited_state=True))
    f_sn = excite(calc, 0, 0, spin=(0, 0))
    prepare_mom_calculation(calc, H2O, f_sn)
    H2O.get_potential_energy()

    dave = Davidson(calc.wfs.eigensolver, None)
    appr_sp_order = dave.estimate_sp_order(calc)
    print(appr_sp_order)

    for kpt in calc.wfs.kpt_u:
        f_sn[kpt.s] = kpt.f_n
    calc.set(eigensolver=LCAOETDM(
        partial_diagonalizer={
            'name': 'Davidson', 'logfile': 'test.txt', 'seed': 42,
            'm': 20, 'eps': 5e-3, 'remember_sp_order': True,
            'sp_order': appr_sp_order},
        linesearch_algo={'name': 'max-step'},
        searchdir_algo={'name': 'LBFGS-P_GMF'},
        localizationtype='PM',
        functional={'name': 'PZ-SIC',
                    'scaling_factor': (0.5, 0.5)},
        need_init_orbs=False),
        occupations={'name': 'mom', 'numbers': f_sn,
                     'use_fixed_occupations': True})

    e = H2O.get_potential_energy()
    assert e == pytest.approx(-2.007241, abs=1.0e-3)

    f = H2O.get_forces()

    f_num = np.array([[-8.01206297e+00, -1.51553367e+01, 3.60670227e-03],
                      [1.42287594e+01, -9.81724693e-01, -5.09333905e-04],
                      [-4.92299436e+00, 1.55306540e+01, 2.12438557e-03]])

    numeric = False
    if numeric:
        from ase.calculators.test import numeric_force
        f_num = np.array([[numeric_force(H2O, a, i)
                          for i in range(3)]
                         for a in range(len(H2O))])
        print('Numerical forces')
        print(f_num)
        print(f - f_num, np.abs(f - f_num).max())

    assert f == pytest.approx(f_num, abs=0.75)



"""Test GW band-gaps for Si."""

import pytest
from ase.build import bulk

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.response.g0w0 import G0W0


def generate_si_systems():
    a = 5.43
    si1 = bulk('Si', 'diamond', a=a)
    si2 = si1.copy()
    si2.positions -= a / 8

    return [si1, si2]


def run(gpw_filename, nblocks):
    # This tests checks the actual numerical accuracy which is asserted below
    calc = GPAW(gpw_filename)
    e = calc.get_potential_energy()

    gw = G0W0(gpw_filename, 'gw_None',
              nbands=8, integrate_gamma=0,
              kpts=[(0, 0, 0), (0.5, 0.5, 0)],  # Gamma, X
              ecut=40, nblocks=nblocks,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1, 'omegamax': None},
              eta=0.2, relbands=(-1, 2))
    results = gw.calculate()

    G, X = results['eps'][0]
    output = [e, G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]]
    G, X = results['qp'][0]
    output += [G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]]

    return output


reference = pytest.approx([-9.253, 5.442, 2.389, 0.403, 0.000,
                           6.261, 3.570, 1.323, 0.001], abs=0.0035)


@pytest.mark.response
@pytest.mark.slow
@pytest.mark.parametrize('si', [0, 1])
@pytest.mark.parametrize('symm', ['all', 'no', 'tr', 'pg'])
@pytest.mark.parametrize('nblocks',
                         [x for x in [1, 2, 4, 8] if x <= world.size])
def test_response_gwsi(in_tmp_dir, si, symm, nblocks, scalapack,
                       gpw_files, gpaw_new):
    if gpaw_new and world.size > 1:
        pytest.skip('Hybrids not working in parallel with GPAW_NEW=1')
    filename = gpw_files[f'si_gw_a{si}_{symm}']
    assert run(filename, nblocks) == reference


@pytest.mark.response
@pytest.mark.ci
@pytest.mark.parametrize('si', [0, 1])
@pytest.mark.parametrize('symm', ['all'])
def test_small_response_gwsi(in_tmp_dir, si, symm, scalapack,
                             gpw_files, gpaw_new):
    if gpaw_new and world.size > 1:
        pytest.skip('Hybrids not working in parallel with GPAW_NEW=1')
    filename = gpw_files[f'si_gw_a{si}_{symm}']
    assert run(filename, 1) == reference


@pytest.mark.response
@pytest.mark.ci
def test_few_freq_response_gwsi(in_tmp_dir, scalapack,
                                gpw_files, gpaw_new):
    if gpaw_new and world.size > 1:
        pytest.skip('Hybrids not working in parallel with GPAW_NEW=1')

    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # This test has very few frequencies and tests that the code doesn't crash.
    filename = gpw_files['si_gw_a0_all']
    gw = G0W0(filename, 'gw_0.2',
              nbands=8, integrate_gamma=0,
              kpts=[(0, 0, 0), (0.5, 0.5, 0)],  # Gamma, X
              ecut=40, nblocks=nblocks,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1, 'omegamax': 0.2},
              eta=0.2, relbands=(-1, 2))
    gw.calculate()



import pytest
from ase import Atoms
from gpaw import GPAW, PW
from gpaw.pipekmezey.pipek_mezey_wannier import PipekMezey


@pytest.mark.pipekmezey
def test_pipekmezey_pw(in_tmp_dir):

    atoms = Atoms('CO',
                  positions=[[0, 0, 0],
                             [0, 0, 1.128]])
    atoms.center(vacuum=5)

    calc = GPAW(mode=PW(200),
                h=0.24,
                convergence={'density': 1e-4,
                             'eigenstates': 1e-4})

    calc.atoms = atoms
    calc.calculate()

    PM = PipekMezey(calc=calc, seed=42)
    PM.localize()

    P = PM.get_function_value()

    assert P == pytest.approx(3.3138, abs=0.0006)
